# coding=utf-8

from grp import getgrgid
from grp import getgrnam
import os
from pwd import getpwnam
from pwd import getpwuid
import shutil
import stat
from typing import Union

from filelock import FileLock


class stfile:
    '''File attributes and permissions

    Manage file owner, group and permissions.
    '''

    def __init__(self, path: str):
        self.__path: str = path

    @property
    def path(self) -> str:
        return self.__path

    @property
    def stat(self) -> os.stat_result:
        '''file stat
        '''
        return os.stat(self.path)

    @property
    def uid(self) -> int:
        '''file uid
        '''
        return self.stat.st_uid

    @uid.setter
    def uid(self, uid: int):
        '''change file uid
        '''
        os.chown(self.path, uid, -1)

    @property
    def gid(self) -> int:
        '''file gid
        '''
        return self.stat.st_gid

    @gid.setter
    def gid(self, gid: int):
        '''change file gid
        '''
        os.chown(self.path, -1, gid)

    @property
    def username(self) -> str:
        '''file owner
        '''
        try:
            return getpwuid(self.uid).pw_name
        except KeyError:  # pragma: no cover
            return str(self.uid)  # pragma: no cover

    @username.setter
    def username(self, owner: Union[int, str]):
        '''change file owner
        '''
        self.uid = getpwnam(owner).pw_uid if isinstance(owner, str) else owner

    @property
    def groupname(self) -> str:
        '''file group
        '''
        try:
            return getgrgid(self.gid).gr_name
        except KeyError:  # pragma: no cover
            return str(self.gid)  # pragma: no cover

    @groupname.setter
    def groupname(self, group: Union[int, str]):
        '''change file group
        '''
        self.gid = getgrnam(group).gr_gid if isinstance(group, str) else group

    def chown(self, owner: Union[int, str], group: Union[int, str] = -1):
        '''change file owner and group
        '''
        self.username = int(owner) if isinstance(owner, str) and owner.isdigit() else owner  # noqa:E501
        self.groupname = int(group) if isinstance(group, str) and group.isdigit() else group  # noqa:E501

    def chgrp(self, group: Union[int, str]):
        '''change group ownership
        '''
        self.groupname = int(group) if isinstance(group, str) and group.isdigit() else group  # noqa:E501

    @property
    def mode(self) -> str:
        '''file mode bits

        6 bits octal digits, e.g. 100644, 040755:

            040764 (drwxrw-r--)
            ┣┛┃┃┃┃
            ┃ ┃┃┃┗━━ Other  ━┓
            ┃ ┃┃┗━━━ Group   ┣━ Permissions
            ┃ ┃┗━━━━ Owner  ━┛
            ┃ ┗━━━━━ Special (setuid, setgid, sticky) Permissions
            ┗━━━━━━━ File type

        File type:

            | 0o140000 | S_IFSOCK | socket            |
            | 0o120000 | S_IFLNK  | symbolic link     |
            | 0o100000 | S_IFREG  | regular file      |
            | 0o060000 | S_IFBLK  | block device      |
            | 0o040000 | S_IFDIR  | directory         |
            | 0o020000 | S_IFCHR  | character device  |
            | 0o010000 | S_IFIFO  | FIFO              |

        Permissions:

            | 7 | rwx | read, write, execute            |
            | 6 | rw- | read, write                     |
            | 5 | r-x | read, execute                   |
            | 4 | r-- | read                            |
            | 3 | -wx | write, execute                  |
            | 2 | -w- | write                           |
            | 1 | --x | execute                         |
            | 0 | --- | no read, no write, no execute   |

        Example:
            stfile("example.txt").mode # 100644
            stfile("example.txt").mode[-4:] # 0755
            stfile("example.txt").mode[-3:] # 764
        '''
        return oct(self.stat.st_mode)[2:].rjust(6, "0")

    @property
    def human_mode(self) -> str:
        '''file mode in `ls -l` format

        10 bits characters, e.g. -rw-r--r--, lrw-rw----, drwxrwxrwx:

             ┏━━━━━━━━━━ Read permission: -/r
             ┃┏━━━━━━━━━ Write permission: -/w
             ┃┃┏━━━━━━━━ Execute permission: -/x/s(x + SUID)/S(only SUID)
             ┃┃┃
             ┃┃┃┏━━━━━━━ Read permission: -/r
             ┃┃┃┃┏━━━━━━ Write permission: -/w
             ┃┃┃┃┃┏━━━━━ Execute permission: -/x/s(x + SGID)/S(only SGID)
             ┃┃┃┃┃┃
             ┃┃┃┃┃┃┏━━━━ Read permission: -/r
             ┃┃┃┃┃┃┃┏━━━ Write permission: -/w
             ┃┃┃┃┃┃┃┃┏━━ Execute permission: -/x/t(x + SBIT)/T(only SBIT)
            -rwxr-xr-x (100755)
            ┃┗┳┛┗┳┛┗┳┛
            ┃ ┃  ┃  ┗━━━ Other  ━┓
            ┃ ┃  ┗━━━━━━ Group   ┣━ Permissions
            ┃ ┗━━━━━━━━━ Owner  ━┛
            ┗━━━━━━━━━━━ File type
        '''
        return stat.filemode(self.stat.st_mode)

    @property
    def human_file_type(self) -> str:
        '''File type

        4 (13 - 16 bits) file type bitmask:

            | - | regular file      | stat.S_ISREG  | S_IFREG   | 0o100000 |
            | d | directory         | stat.S_ISDIR  | S_IFDIR   | 0o040000 |
            | l | symbolic link     | stat.S_ISLNK  | S_IFLNK   | 0o120000 |
            | b | block device      | stat.S_ISBLK  | S_IFBLK   | 0o060000 |
            | c | character device  | stat.S_ISCHR  | S_IFCHR   | 0o020000 |
            | s | socket            | stat.S_ISSOCK | S_IFSOCK  | 0o140000 |
            | p | FIFO              | stat.S_ISFIFO | S_IFIFO   | 0o010000 |
        '''
        return self.human_mode[0]

    @property
    def human_all_permissions(self) -> str:
        '''File all users permissions

        9 (1 - 9 bits) users permissions bitmask and
        3 (10 - 12 bits) special permissions
        '''
        return self.human_mode[-9:]

    @property
    def human_owner_permissions(self) -> str:
        '''File owner permissions

        3 (7 - 9 bits) owner permissions bitmask and setuid bit:
            - Read permission: -/r
            - Write permission: -/w
            - Execute permission: -/x/s(x + SUID)/S(only SUID)
        '''
        return self.human_all_permissions[0:3]

    @property
    def human_group_permissions(self) -> str:
        '''File group permissions

        3 (4 - 6 bits) group permissions bitmask and setgid bit:
            - Read permission: -/r
            - Write permission: -/w
            - Execute permission: -/x/s(x + SGID)/S(only SGID)
        '''
        return self.human_all_permissions[3:6]

    @property
    def human_other_permissions(self) -> str:
        '''File other (not in group) permissions

        3 (1 - 3 bits) other permissions bitmask and sticky bit:
            - Read permission: -/r
            - Write permission: -/w
            - Execute permission: -/x/t(x + SBIT)/T(only SBIT)
        '''
        return self.human_all_permissions[6:9]

    def chmod(self, mode: Union[int, str]):
        '''change file mode bits
        '''
        if isinstance(mode, str):
            mode = int(mode, 8)
        os.chmod(self.path, mode)


class safile:
    '''Secure read and write files

    Backup before writing and restore (if backup exists) before reading.
    '''

    @classmethod
    def lock(cls, origin: str):
        '''Unified file lock
        '''
        return FileLock(f"{origin}.lock")

    @classmethod
    def get_backup_path(cls, origin: str) -> str:
        '''Unified backup path
        '''
        return f"{origin}.bak"

    @classmethod
    def create_backup(cls, path: str, copy: bool = False) -> bool:
        '''Create a backup before writing file

        Backup files with '.bak' suffix will be created in the same directory.
        By default shutil.move() is used to create the backup file, which will
        use os.rename() to rename the original file. This will make the backup
        very efficient.
        But, if you wish to append to the original file, you need to specify
        'copy=True' to use shutil.copy2().
        '''
        pbak: str = cls.get_backup_path(path)
        if os.path.isfile(pbak):  # Restore before creating a new backup
            assert cls.restore(path), f"restore '{path}' failed"
        assert not os.path.exists(pbak), f"backup file '{pbak}' already exists"
        if not os.path.exists(path):  # No need for backup
            return True
        assert os.path.isfile(path), f"'{path}' is not a regular file"
        method = shutil.copy2 if copy else shutil.move
        assert method(src=path, dst=pbak) == pbak, f"backup '{path}' failed"
        return os.path.exists(pbak)

    @classmethod
    def delete_backup(cls, path: str) -> bool:
        '''Delete backup after writing file
        '''
        pbak: str = cls.get_backup_path(path)
        if os.path.isfile(pbak):
            os.remove(pbak)
        return not os.path.exists(pbak)

    @classmethod
    def restore(cls, path: str) -> bool:
        '''Restore (if backup exists) before reading file
        '''
        pbak: str = cls.get_backup_path(path)
        if os.path.isfile(pbak):
            if os.path.isfile(path):
                os.remove(path)
            assert not os.path.exists(path), f"file '{path}' still exists"
            assert shutil.move(src=pbak, dst=path) == path, \
                f"restore backup file '{pbak}' to '{path}' failed"
        return not os.path.exists(pbak)



# coding=utf-8

import os
from typing import List


def singleton(cls):
    instance = {}

    def _singleton_wrapper(*args, **kwargs):
        if cls not in instance:
            instance[cls] = cls(*args, **kwargs)
        return instance[cls]

    return _singleton_wrapper


@singleton
class chdir:
    '''Change working directory
    '''

    def __init__(self):
        self.__stack: List[str] = []

    def pushd(self, path: str):
        '''Add directory to stack.
        '''
        assert isinstance(path, str), f"type '{type(path)}' is not str"
        assert os.path.isdir(path), f"path '{path}' is not directory"
        self.__stack.append(os.getcwd())
        os.chdir(path)

    def popd(self):
        '''Remove directory from stack.
        '''
        assert len(self.__stack) > 0
        os.chdir(self.__stack.pop())



# coding:utf-8

from concurrent.futures import ThreadPoolExecutor
from queue import Queue
import sys
from threading import Lock
from threading import Thread
from threading import current_thread  # noqa:H306
from time import time
from typing import Any
from typing import Callable
from typing import Dict
from typing import Generic
from typing import Iterator
from typing import Optional
from typing import Set
from typing import Tuple
from typing import TypeVar

from xkits.actuator import Logger
from xkits.actuator import commands  # noqa:H306

LKIT = TypeVar("LKIT")
LKNT = TypeVar("LKNT")


class NamedLock(Generic[LKNT]):

    class LockItem(Generic[LKIT]):
        def __init__(self, name: LKIT):
            self.__lock: Lock = Lock()
            self.__name: LKIT = name

        @property
        def name(self) -> LKIT:
            return self.__name

        @property
        def lock(self) -> Lock:
            return self.__lock

    def __init__(self):
        self.__locks: Dict[LKNT, NamedLock.LockItem[LKNT]] = {}
        self.__inter: Lock = Lock()  # internal lock

    def __len__(self) -> int:
        return len(self.__locks)

    def __iter__(self) -> Iterator[LockItem[LKNT]]:
        return iter(self.__locks.values())

    def __contains__(self, name: LKNT) -> bool:
        return name in self.__locks

    def __getitem__(self, name: LKNT) -> Lock:
        return self.lookup(name).lock

    def lookup(self, name: LKNT) -> LockItem[LKNT]:
        try:
            return self.__locks[name]
        except KeyError:
            with self.__inter:
                if name not in self.__locks:
                    lock = self.LockItem(name)
                    self.__locks.setdefault(name, lock)
                    assert self.__locks[name] is lock
                    return lock

                lock = self.__locks[name]  # pragma: no cover
                assert lock.name == name  # pragma: no cover
                return lock  # pragma: no cover


class ThreadPool(ThreadPoolExecutor):
    '''Thread Pool'''

    def __init__(self, max_workers: Optional[int] = None,
                 thread_name_prefix: str = "work_thread",
                 initializer: Optional[Callable] = None,
                 initargs: Tuple = ()):
        '''Initializes an instance based on ThreadPoolExecutor.'''
        self.__cmds: commands = commands()
        if isinstance(max_workers, int):
            max_workers = max(max_workers, 2)
        super().__init__(max_workers, thread_name_prefix, initializer, initargs)  # noqa:E501

    @property
    def cmds(self) -> commands:
        '''command-line toolkit'''
        return self.__cmds

    @property
    def alive_threads(self) -> Set[Thread]:
        '''alive threads'''
        return {thread for thread in self._threads if thread.is_alive()}

    @property
    def other_threads(self) -> Set[Thread]:
        '''other threads'''
        current: Thread = current_thread()
        return {thread for thread in self._threads if thread is not current}

    @property
    def other_alive_threads(self) -> Set[Thread]:
        '''other alive threads'''
        return {thread for thread in self.other_threads if thread.is_alive()}


class TaskJob():  # pylint: disable=too-many-instance-attributes
    '''Task Job'''

    def __init__(self, no: int, fn: Callable, *args: Any, **kwargs: Any):
        self.__no: int = no
        self.__fn: Callable = fn
        self.__args: Tuple[Any, ...] = args
        self.__kwargs: Dict[str, Any] = kwargs
        self.__result: Any = LookupError(f"Job{no} is not started")
        self.__created: float = time()
        self.__started: float = 0.0
        self.__stopped: float = 0.0

    def __str__(self) -> str:
        args = list(self.args) + list(f"{k}={v}" for k, v in self.kwargs)
        info: str = ", ".join(f"{a}" for a in args)
        return f"Job{self.id} {self.fn}({info})"

    @property
    def id(self) -> int:
        '''job id'''
        return self.__no

    @property
    def fn(self) -> Callable:
        '''job callable function'''
        return self.__fn

    @property
    def args(self) -> Tuple[Any, ...]:
        '''job callable arguments'''
        return self.__args

    @property
    def kwargs(self) -> Dict[str, Any]:
        '''job callable keyword arguments'''
        return self.__kwargs

    @property
    def result(self) -> Any:
        '''job callable function return value'''
        if isinstance(self.__result, Exception):
            raise self.__result
        return self.__result

    @property
    def created(self) -> float:
        '''job created time'''
        return self.__created

    @property
    def started(self) -> float:
        '''job started time'''
        return self.__started

    @property
    def stopped(self) -> float:
        '''job stopped time'''
        return self.__stopped

    def run(self) -> bool:
        '''run job'''
        try:
            if self.__started > 0.0:
                raise RuntimeError(f"{self} is already started")
            self.__started = time()
            self.__result = self.fn(*self.args, **self.kwargs)
            return True
        except Exception as error:  # pylint: disable=broad-exception-caught
            self.__result = error
            return False
        finally:
            self.__stopped = time()


if sys.version_info >= (3, 9):
    JobQueue = Queue[Optional[TaskJob]]  # noqa: E501, pragma: no cover, pylint: disable=unsubscriptable-object
else:  # Python3.8 TypeError
    JobQueue = Queue  # pragma: no cover


class TaskPool(Dict[int, TaskJob]):  # noqa: E501, pylint: disable=too-many-instance-attributes
    '''Task Thread Pool'''

    def __init__(self, workers: int = 1, jobs: int = 0, prefix: str = "task"):
        wsize: int = max(workers, 1)
        qsize = max(wsize, jobs) if jobs > 0 else jobs
        self.__cmds: commands = commands()
        self.__jobs: JobQueue = Queue(qsize)
        self.__prefix: str = prefix or "task"
        self.__threads: Set[Thread] = set()
        self.__intlock: Lock = Lock()  # internal lock
        self.__running: bool = False
        self.__workers: int = wsize
        self.__counter: int = 0
        self.__suceess: int = 0
        self.__failure: int = 0
        super().__init__()

    def __enter__(self):
        self.startup()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.shutdown()

    @property
    def jobs(self) -> JobQueue:
        '''task jobs'''
        return self.__jobs

    @property
    def cmds(self) -> commands:
        '''command-line toolkit'''
        return self.__cmds

    @property
    def thread_name_prefix(self) -> str:
        '''task thread name prefix'''
        return self.__prefix

    @property
    def threads(self) -> Set[Thread]:
        '''task threads'''
        return self.__threads

    @property
    def running(self) -> bool:
        '''task threads are started'''
        return self.__running

    @property
    def workers(self) -> int:
        '''task workers'''
        return self.__workers

    @property
    def counter(self) -> int:
        '''task job counter'''
        return self.__counter

    @property
    def suceess(self) -> int:
        '''suceess job counter'''
        return self.__suceess

    @property
    def failure(self) -> int:
        '''suceess job counter'''
        return self.__failure

    def task(self):
        '''execute a task from jobs queue'''
        counter: int = 0
        suceess: int = 0
        failure: int = 0
        logger: Logger = self.cmds.logger
        logger.debug("Task thread %s is running", current_thread().name)
        while True:
            job: Optional[TaskJob] = self.jobs.get(block=True)
            if job is None:  # stop task
                self.jobs.put(job)  # notice other tasks
                break
            counter += 1
            if not job.run():
                self.__failure += 1
                failure += 1
            else:
                self.__suceess += 1
                suceess += 1
        logger.debug("Task thread %s is stopped, %s", current_thread().name,
                     f"{counter} jobs: {suceess} suceess and {failure} failure"
                     )

    def submit(self, fn: Callable, *args: Any, **kwargs: Any) -> TaskJob:
        '''submit a task to jobs queue

        Returns:
            int: job id
        '''
        sn: int  # serial number
        with self.__intlock:  # generate job id under lock protection
            self.__counter += 1
            sn = self.__counter
        job: TaskJob = TaskJob(sn, fn, *args, **kwargs)
        self.jobs.put(job, block=True)
        self.setdefault(sn, job)
        assert self[sn] is job
        return job

    def shutdown(self) -> None:
        '''stop all task threads and waiting for all jobs finish'''
        with self.__intlock:  # block submit new tasks
            self.cmds.logger.debug("Shutdown %s tasks", self.thread_name_prefix)  # noqa:E501
            self.__running = False
            self.jobs.put(None)  # notice tasks
            while len(self.threads) > 0:
                thread: Thread = self.threads.pop()
                thread.join()
            while not self.jobs.empty():
                job: Optional[TaskJob] = self.jobs.get(block=True)
                if job is not None:  # shutdown only after executed
                    raise RuntimeError(f"Unexecuted job: {job}")  # noqa:E501, pragma: no cover

    def startup(self) -> None:
        '''start task threads'''
        with self.__intlock:
            self.cmds.logger.debug("Startup %s tasks", self.thread_name_prefix)
            for i in range(self.workers):
                thread_name: str = f"{self.thread_name_prefix}_{i}"
                thread = Thread(name=thread_name, target=self.task)
                self.threads.add(thread)
                thread.start()  # run
            self.__running = True

    def barrier(self) -> None:
        '''stop submit new tasks and waiting for all submitted tasks to end'''
        self.shutdown()
        self.startup()



# coding:utf-8

import os
from typing import Optional
from typing import Sequence

from xkits.actuator import add_command
from xkits.actuator import commands
from xkits.actuator import run_command
from xkits.attribute import __author__
from xkits.attribute import __author_email__
from xkits.attribute import __prog_name__
from xkits.attribute import __prog_project__
from xkits.attribute import __project__
from xkits.attribute import __urlhome__
from xkits.attribute import __version__
from xkits.parser import argp


class project:

    def __init__(self, name: str, license: str, allow_update: bool = False):  # pylint: disable=redefined-builtin # noqa:E501
        # check illegal characters in project name
        for char in [" "]:
            if char in name:
                raise ValueError(f"Illegal character '{char}' in '{name}'")
        folder: str = name.replace("-", "_")
        self.__name: str = name
        self.__folder: str = folder
        self.__license: str = license
        self.__allow_update: bool = allow_update

    @property
    def name(self) -> str:
        return self.__name

    @property
    def folder(self) -> str:
        return self.__folder

    @property
    def license(self) -> str:
        return self.__license

    @property
    def allow_update(self) -> bool:
        return self.__allow_update

    def write(self, path: str, content: str) -> bool:
        if not os.path.exists(path) or self.allow_update:
            with open(path, "w", encoding="utf-8") as whdl:
                if len(content) > 0 and content[-1] != "\n":
                    content += "\n"
                whdl.write(content)
        return True

    def init_requirements(self):
        self.write("requirements.txt", f'''{__project__} >= {__version__}''')

    def init_coveragerc(self):
        self.write(".coveragerc", f'''[run]
omit =
    {self.folder}/unittest/*
    {self.folder}/attribute.py

[report]
exclude_lines =
    pragma: no cover
    raise NotImplementedError
    if __name__ == .__main__.:
    def __repr__
''')

    def init_pylintrc(self):
        self.write(".pylintrc", '''[MASTER]
disable=
    C0103,   # invalid-name
    C0114,   # missing-module-docstring
    C0115,   # missing-class-docstring
    C0116,   # missing-function-docstring
    C0301,   # line-too-long
''')

    def init_makefile(self):
        self.write("Makefile", f'''MAKEFLAGS += --always-make

all: build reinstall test


clean-cover:
	rm -rf cover .coverage coverage.xml htmlcov
clean-tox:
	rm -rf .stestr .tox
clean: build-clean test-clean clean-cover clean-tox


upload:
	xpip-upload --config-file .pypirc dist/*


build-clean:
	xpip-build --debug setup --clean
build-requirements:
	pip3 install -r requirements.txt
build: build-clean build-requirements
	xpip-build --debug setup --all


install:
	pip3 install --force-reinstall --no-deps dist/*.whl
uninstall:
	pip3 uninstall -y {self.name}
reinstall: uninstall install


test-prepare:
	pip3 install --upgrade mock pylint flake8 pytest pytest-cov
pylint:
	pylint $(shell git ls-files {self.folder}/*.py test/*.py example/*.py)
flake8:
	flake8 {self.folder} --count --select=E9,F63,F7,F82 --show-source --statistics
	flake8 {self.folder} --count --exit-zero --max-complexity=10 --max-line-length=127 --statistics
pytest:
	pytest --cov={self.folder} --cov-report=term-missing --cov-report=xml --cov-report=html --cov-config=.coveragerc --cov-fail-under=100
pytest-clean:
	rm -rf .pytest_cache
test: test-prepare pylint flake8 pytest
test-clean: pytest-clean
''')  # noqa:W191,E101,E501

    def init_project(self):
        os.makedirs(self.folder, exist_ok=True)
        os.makedirs(os.path.join(self.folder, "unittest"), exist_ok=True)
        self.write(os.path.join(self.folder, "__init__.py"), "")
        self.write(os.path.join(self.folder, "unittest", "__init__.py"), "")
        self.write(os.path.join(self.folder, "attribute.py"),
                   f'''# coding:utf-8

from urllib.parse import urljoin

__project__ = "{self.name}"
__version__ = "0.1.alpha.1"
__description__ = "Automatically created by {__prog_project__}."
__urlhome__ = "{__urlhome__}"
__urlcode__ = __urlhome__
__urldocs__ = __urlhome__
__urlbugs__ = urljoin(__urlhome__, "issues")

# author
__author__ = "{__author__}"
__author_email__ = "{__author_email__}"
''')
        self.write(os.path.join(self.folder, "command.py"),
                   f'''# coding:utf-8

from typing import Optional
from typing import Sequence

from {__prog_name__} import add_command
from {__prog_name__} import argp
from {__prog_name__} import commands
from {__prog_name__} import run_command

from {self.folder}.attribute import __description__
from {self.folder}.attribute import __project__
from {self.folder}.attribute import __urlhome__
from {self.folder}.attribute import __version__


@add_command(__project__, description=__description__)
def add_cmd(_arg: argp):
    pass


@run_command(add_cmd)
def run_cmd(cmds: commands) -> int:
    return 0


def main(argv: Optional[Sequence[str]] = None) -> int:
    cmds = commands()
    cmds.version = __version__
    return cmds.run(root=add_cmd, argv=argv, epilog=f"For more, please visit {{__urlhome__}}.")  # noqa:E501
''')

    def init_readme(self):
        self.write("README.md", f'''# {self.name}

> Automatically created by {__prog_project__}.''')

    def init_setup(self):
        # create setup.cfg
        self.write("setup.cfg", f'''[metadata]
keywords = command-line, argparse, argcomplete
long_description = file: README.md
long_description_content_type = text/markdown
license = {self.license}
license_files = LICENSE
platforms = any
classifiers =
    Programming Language :: Python
    Programming Language :: Python :: 3

[options]
zip_safe = True
include_package_data = True
python_requires = >=3.8

[options.entry_points]
console_scripts =
    {self.name} = {self.folder}.command:main
''')

        # create setup.py
        self.write("setup.py", f'''# coding=utf-8

from setuptools import find_packages
from setuptools import setup

from {self.folder}.attribute import __author__
from {self.folder}.attribute import __author_email__
from {self.folder}.attribute import __description__
from {self.folder}.attribute import __project__
from {self.folder}.attribute import __urlbugs__
from {self.folder}.attribute import __urlcode__
from {self.folder}.attribute import __urldocs__
from {self.folder}.attribute import __urlhome__
from {self.folder}.attribute import __version__


def all_requirements():
    def read_requirements(path: str):
        with open(path, "r", encoding="utf-8") as rhdl:
            return rhdl.read().splitlines()

    requirements = read_requirements("requirements.txt")
    return requirements


setup(
    name=__project__,
    version=__version__,
    description=__description__,
    url=__urlhome__,
    author=__author__,
    author_email=__author_email__,
    project_urls={{"Source Code": __urlcode__,
                  "Bug Tracker": __urlbugs__,
                  "Documentation": __urldocs__}},
    packages=find_packages(include=["{self.folder}*"], exclude=["{self.folder}.unittest"]),
    install_requires=all_requirements())
''')  # noqa:E501

    def create(self) -> int:
        self.init_requirements()
        self.init_coveragerc()
        self.init_pylintrc()
        self.init_makefile()
        self.init_project()
        self.init_readme()
        self.init_setup()
        return 0


@add_command("init", help="Initialize a command-line project.")
def add_cmd_init(_arg: argp):
    _arg.add_opt_on("--update", help="allow updating existing files")
    _arg.add_argument("--license", type=str, nargs=1, metavar="LICENSE",
                      default=["MIT"], help="select license, default to MIT")
    _arg.add_pos("project_name", type=str, metavar="PROJECT")


@run_command(add_cmd_init)
def run_cmd_init(cmds: commands) -> int:
    return project(name=cmds.args.project_name,
                   license=cmds.args.license[0],
                   allow_update=cmds.args.update
                   ).create()


@add_command(__prog_project__, description="Create a command-line project.")
def add_cmd(_arg: argp):
    pass


@run_command(add_cmd, add_cmd_init)
def run_cmd(cmds: commands) -> int:  # pylint: disable=unused-argument
    return 0


def main(argv: Optional[Sequence[str]] = None) -> int:
    cmds = commands()
    cmds.version = __version__
    return cmds.run(root=add_cmd, argv=argv, epilog=f"For more, please visit {__urlhome__}.")  # noqa:E501



# coding:utf-8

from argparse import Namespace
from errno import ECANCELED
from errno import EINVAL
from errno import ENOENT
from errno import ENOTRECOVERABLE
import logging
from logging import Logger
import sys
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple

from xkits.attribute import __prog_name__
from xkits.colorful import color
from xkits.logger import log
from xkits.parser import argp
from xkits.utils import singleton


class add_command:
    '''Define a new command-line node.

    For example:

    >>> from xkits import add_command\n
    >>> from xkits import argp\n

    >>> @add_command("example")\n
    >>> def cmd(_arg: argp):\n
    >>>     _arg.add_opt_on("-t", "--test")\n
    '''

    def __init__(self, name: str, **kwargs):
        '''
        @param name: Node name
        @type name: str

        @param description: Text to display before the argument help
        @type description: str (by default, no text)

        @param epilog: Text to display after the argument help
        @type epilog: str (by default, no text)

        @param help: Help message as a subcommand
        @type help: str

        @param add_help: Add a -h/--help option to the node
        @type add_help: bool (default: True)
        '''
        if "help" in kwargs and "description" not in kwargs:
            kwargs["description"] = kwargs["help"]
        if "description" in kwargs and "help" not in kwargs:
            kwargs["help"] = kwargs["description"]
        self.__name: str = name
        self.__prev: add_command = self
        self.__cmds: commands = commands()
        self.__options: Dict[str, Any] = kwargs
        self.__bind: Optional[run_command] = None
        self.__subs: Optional[Tuple[add_command, ...]] = None
        self.__func: Optional[Callable[[argp], None]] = None

    def __call__(self, cmd_func: Callable[[argp], None]):
        self.__func = cmd_func
        return self

    @property
    def func(self) -> Callable[[argp], None]:
        if self.__func is None:
            raise ValueError("No function")  # pragma: no cover
        return self.__func

    @property
    def name(self) -> str:
        return self.__name

    @property
    def root(self) -> "add_command":
        root = self.__prev
        while root.prev != root:
            root = root.prev
        return root

    @property
    def prev(self) -> "add_command":
        return self.__prev

    @prev.setter
    def prev(self, value: "add_command"):
        assert isinstance(value, add_command)
        self.__prev = value

    @property
    def cmds(self) -> "commands":
        return self.__cmds

    @property
    def options(self) -> Dict[str, Any]:
        return self.__options

    @property
    def bind(self) -> Optional["run_command"]:
        return self.__bind

    @bind.setter
    def bind(self, value: "run_command"):
        assert isinstance(value, run_command)
        self.__bind = value

    @property
    def subs(self) -> Optional[Tuple["add_command", ...]]:
        return self.__subs

    @subs.setter
    def subs(self, value: Tuple["add_command", ...]):
        assert isinstance(value, Tuple)
        for sub in value:
            assert isinstance(sub, add_command)
        self.__subs = value

    @property
    def sub_dest(self) -> str:
        node: add_command = self
        subs: List[str] = [self.name]
        while node.prev is not node:
            node = node.prev
            subs.insert(0, node.name)
        name = "_".join(subs)
        return f"__sub_dest_{name}__"


class run_command:
    '''Define the main callback function, and bind it to a node and
    all subcommands.

    For example:

    >>> from xkits import commands\n
    >>> from xkits import run_command\n

    >>> @run_command(cmd, cmd_get, cmd_set)\n
    >>> def run(cmds: commands) -> int:\n
    >>>     return 0\n
    '''

    def __init__(self, cmd_bind: add_command, *sub_cmds: add_command,
                 skip: bool = False):
        '''
        @param cmd_bind: Bind to a root command node
        @type name: add_command

        @param *sub_cmds: All required subcommands
        @type *sub_cmds: add_command

        @param skip: This node (run_command, pre_command and end_command)
        does not run when a subcommand is specified, run this node without
        any subcommands
        @type skip: bool (default: False)
        '''
        assert isinstance(cmd_bind, add_command)
        assert isinstance(skip, bool)
        cmd_bind.bind = self
        cmd_bind.subs = sub_cmds
        for sub in sub_cmds:
            sub.prev = cmd_bind
        commands().root = cmd_bind.root
        self.__skip: bool = skip
        self.__bind: add_command = cmd_bind
        self.__prep: Optional["pre_command"] = None
        self.__done: Optional["end_command"] = None
        self.__func: Optional[Callable[["commands"], int]] = None

    def __call__(self, run_func: Callable[["commands"], int]):
        self.__func = run_func
        return self

    @property
    def func(self) -> Callable[["commands"], int]:
        if self.__func is None:
            raise ValueError("No function")  # pragma: no cover
        return self.__func

    @property
    def bind(self) -> add_command:
        return self.__bind

    @property
    def prep(self) -> Optional["pre_command"]:
        return self.__prep

    @prep.setter
    def prep(self, value: "pre_command"):
        assert isinstance(value, pre_command)
        self.__prep = value

    @property
    def done(self) -> Optional["end_command"]:
        return self.__done

    @done.setter
    def done(self, value: "end_command"):
        assert isinstance(value, end_command)
        self.__done = value

    @property
    def skip(self) -> bool:
        return self.__skip


class pre_command:
    '''Define prepare callback function, and bind it with main callback.

    For example:

    >>> from xkits import commands\n
    >>> from xkits import pre_command\n
    >>> from xkits import run_command\n

    >>> @run_command(cmd, cmd_get, cmd_set)\n
    >>> def run(cmds: commands) -> int:\n
    >>>     return 0\n

    >>> @pre_command(run)\n
    >>> def pre(cmds: commands) -> int:\n
    >>>     return 0\n
    '''

    def __init__(self, run_bind: run_command):
        '''
        @param cmd_bind: Bind to a root command node
        @type name: add_command
        '''
        assert isinstance(run_bind, run_command)
        run_bind.prep = self
        self.__main: run_command = run_bind
        self.__func: Optional[Callable[["commands"], int]] = None

    def __call__(self, run_func: Callable[["commands"], int]):
        self.__func = run_func
        return self

    @property
    def func(self) -> Callable[["commands"], int]:
        if self.__func is None:
            raise ValueError("No function")  # pragma: no cover
        return self.__func

    @property
    def main(self) -> run_command:
        return self.__main


class end_command:
    '''Define purge callback function, and bind it with main callback.

    For example:

    >>> from xkits import commands\n
    >>> from xkits import end_command\n
    >>> from xkits import run_command\n

    >>> @run_command(cmd, cmd_get, cmd_set)\n
    >>> def run(cmds: commands) -> int:\n
    >>>     return 0\n

    >>> @end_command(run)\n
    >>> def end(cmds: commands) -> int:\n
    >>>     return 0\n
    '''

    def __init__(self, run_bind: run_command):
        '''
        @param cmd_bind: Bind to a root command node
        @type name: add_command
        '''
        assert isinstance(run_bind, run_command)
        run_bind.done = self
        self.__main: run_command = run_bind
        self.__func: Optional[Callable[["commands"], int]] = None

    def __call__(self, run_func: Callable[["commands"], int]):
        self.__func = run_func
        return self

    @property
    def func(self) -> Callable[["commands"], int]:
        if self.__func is None:
            raise ValueError("No function")  # pragma: no cover
        return self.__func

    @property
    def main(self) -> run_command:
        return self.__main


@singleton
class commands(log):
    '''Singleton command-line tool based on argparse.

    Define and bind all callback functions before calling run() or parse().

    For example:

    >>> from typing import Optional\n
    >>> from typing import Sequence\n

    >>> from xkits import add_command\n
    >>> from xkits import argp\n
    >>> from xkits import commands\n
    >>> from xkits import end_command\n
    >>> from xkits import pre_command\n
    >>> from xkits import run_command\n

    >>> @add_command("example")\n
    >>> def cmd(_arg: argp):\n
    >>>     _arg.add_opt_on("-t", "--test")\n

    >>> @run_command(cmd, cmd_get, cmd_set)\n
    >>> def run(cmds: commands) -> int:\n
    >>>     return 0\n

    >>> @pre_command(run)\n
    >>> def pre(cmds: commands) -> int:\n
    >>>     return 0\n

    >>> @end_command(run)\n
    >>> def end(cmds: commands) -> int:\n
    >>>     return 0\n

    >>> def main(argv: Optional[Sequence[str]] = None) -> int:\n
    >>>     return commands().run(\n
    >>>         root=cmd,\n
    >>>         argv=argv,\n
    >>>         prog="xkits-example",\n
    >>>         description="Simple command-line tool based on argparse.")\n
    '''

    LOGGER_ARGUMENT_GROUP = "logger options"

    def __init__(self):
        self.__prog: str = __prog_name__
        self.__root: Optional[add_command] = None
        self.__args: Namespace = Namespace()
        self.__version: Optional[str] = None
        self.__enabled_logger: bool = True
        super().__init__()

    @property
    def prog(self) -> str:
        return self.__prog

    @property
    def root(self) -> Optional[add_command]:
        '''Root Command.
        '''
        return self.__root

    @root.setter
    def root(self, value: add_command):
        assert isinstance(value, add_command)
        self.__root = value

    @property
    def args(self) -> Namespace:
        '''Namespace after parse arguments.
        '''
        assert isinstance(self.__args, Namespace)
        return self.__args

    @args.setter
    def args(self, value: Namespace):
        assert isinstance(value, Namespace)
        self.__args = value

    @property
    def version(self) -> Optional[str]:
        '''Custom version for "-v" or "--version" output.
        '''
        return self.__version

    @version.setter
    def version(self, value: str):
        assert isinstance(value, str)
        _version = value.strip()
        self.__version = _version

    @property
    def enabled_logger(self) -> bool:
        return self.__enabled_logger

    @enabled_logger.setter
    def enabled_logger(self, value: bool):
        assert isinstance(value, bool)
        self.__enabled_logger = value

    @property
    def logger(self) -> Logger:
        '''Logger.
        '''
        return self.get_logger(self.prog)

    def stdout(self, context: Any):
        '''Output string to sys.stdout.
        '''
        sys.stdout.write(f"{context}\n")
        sys.stdout.flush()

    def stdout_green(self, context: Any):
        '''Output string to sys.stdout with green color.
        '''
        self.stdout(color.green(context))

    def stderr(self, context: Any):
        '''Output string to sys.stderr.
        '''
        sys.stderr.write(f"{context}\n")
        sys.stderr.flush()

    def stderr_red(self, context: Any):
        '''Output string to sys.stderr with red color.
        '''
        self.stderr(color.red(context))

    def __add_optional_version(self, _arg: argp):
        version = self.version
        if not isinstance(version, str):
            return

        options = _arg.filter_optional_name("-v", "--version")
        if len(options) > 0:
            _arg.add_argument(*options, action="version",
                              version=f"%(prog)s {version.strip()}")

    def __add_inner_parser_tail(self, _arg: argp):

        def filter_optional_name(*name: str) -> Optional[str]:
            options = _arg.filter_optional_name(*name)
            if len(options) > 0:
                for i in name:
                    if i in options:
                        return i
            return None

        def add_optional_level():
            group = _arg.argument_group(self.LOGGER_ARGUMENT_GROUP)
            group_level = group.add_mutually_exclusive_group()

            option_level = filter_optional_name("--level", "--log-level")
            if isinstance(option_level, str):
                group_level.add_argument(
                    option_level,
                    type=str,
                    nargs="?",
                    const=self.LOG_LEVELS.INFO.value,
                    default=self.LOG_LEVELS.INFO.value,
                    choices=self.ALLOWED_LOG_LEVELS,
                    dest="_log_level_str_",
                    help="Logger output level, default info.")

            for level in self.ALLOWED_LOG_LEVELS:
                options = []
                if isinstance(filter_optional_name(f"-{level[0]}"), str):
                    options.append(f"-{level[0]}")
                if isinstance(filter_optional_name(f"--{level}"), str):
                    options.append(f"--{level}")
                elif isinstance(filter_optional_name(f"--{level}-level"), str):
                    options.append(f"--{level}-level")

                if not options:
                    continue
                group_level.add_argument(*options,
                                         action="store_const",
                                         const=level,
                                         dest="_log_level_str_",
                                         help=f"Logger level set to {level}.")

        def add_optional_stream():
            option = filter_optional_name("--log", "--log-file")
            if not isinstance(option, str):
                return

            group = _arg.argument_group(self.LOGGER_ARGUMENT_GROUP)
            group.add_argument(option,
                               type=str,
                               nargs=1,
                               default=[],
                               metavar="FILE",
                               action="extend",
                               dest="_log_files_",
                               help="Logger output to file.")

        def add_optional_format():
            option = filter_optional_name("--format", "--log-format")
            if not isinstance(option, str):
                return

            DEFAULT_LOG_FMT = "%(log_color)s%(asctime)s"\
                " %(process)d %(threadName)s %(levelname)s"\
                " %(funcName)s %(filename)s:%(lineno)s"\
                " %(message)s"

            group = _arg.argument_group(self.LOGGER_ARGUMENT_GROUP)
            group.add_argument(option,
                               type=str,
                               nargs="?",
                               const=DEFAULT_LOG_FMT,
                               default=self.DEFAULT_LOG_FORMAT,
                               metavar="STRING",
                               dest="_log_format_",
                               help="Logger output format.")

        def add_optional_console():
            group = _arg.argument_group(self.LOGGER_ARGUMENT_GROUP)
            group_std = group.add_mutually_exclusive_group()

            option = filter_optional_name("--stdout", "--log-stdout")
            if isinstance(option, str):
                group_std.add_argument(option,
                                       const=sys.stdout,
                                       action="store_const",
                                       dest="_log_console_",
                                       help="Logger output to stdout.")

            option = filter_optional_name("--stderr", "--log-stderr")
            if isinstance(option, str):
                group_std.add_argument(option,
                                       const=sys.stderr,
                                       action="store_const",
                                       dest="_log_console_",
                                       help="Logger output to stderr.")

        if self.enabled_logger:
            add_optional_level()
            add_optional_stream()
            add_optional_format()
            add_optional_console()

    def __parse_logger(self, args: Namespace):
        if not self.enabled_logger:
            return

        def parse_format() -> Optional[str]:
            if hasattr(args, "_log_format_"):
                fmt = getattr(args, "_log_format_")
                if isinstance(fmt, str):
                    return fmt
            return None

        def parse_level() -> Optional[str]:
            if hasattr(args, "_log_level_str_"):
                level = getattr(args, "_log_level_str_")
                if isinstance(level, str):
                    return level.upper()
            return None

        def parse_console() -> Optional[Any]:
            return getattr(args, "_log_console_", None)

        def parse_files() -> List[str]:
            return getattr(args, "_log_files_", [])

        fmt: Optional[str] = parse_format()
        level_name: Optional[str] = parse_level()
        console: Optional[Any] = parse_console()

        handlers: List[logging.Handler] = []
        if console is not None:
            handlers.append(log.new_stream_handler(stream=console, fmt=fmt))
        for filename in parse_files():
            handlers.append(log.new_file_handler(filename=filename, fmt=fmt))
        self.initiate_logger(self.logger, level=level_name, handlers=handlers)

    def __add_parser(self, _map: Dict[add_command, argp],
                     arg_root: argp, cmd_root: add_command, **kwargs):
        assert isinstance(cmd_root, add_command)
        assert cmd_root not in _map
        _map[cmd_root] = arg_root

        if not cmd_root.subs or len(cmd_root.subs) <= 0:
            return

        _sub = arg_root.add_subparsers(dest=cmd_root.sub_dest)
        for sub in cmd_root.subs:
            assert isinstance(sub, add_command)
            options = sub.options.copy()
            for key, value in kwargs.items():
                options.setdefault(key, value)
            options.setdefault("epilog", arg_root.epilog)
            options.setdefault("prev_parser", arg_root)
            _arg: argp = _sub.add_parser(sub.name, **options)
            self.__add_parser(_map, _arg, sub)

    def __add_option(self, _map: Dict[add_command, argp]):
        for _cmd, _arg in _map.items():
            _cmd.func(_arg)
            self.__add_inner_parser_tail(_arg)

    @classmethod
    def check_error(cls, value: Any) -> int:
        '''Check value is an error.

        Return True if value is an error, otherwise False.
        '''
        return value if isinstance(value, int) else 0 if value in (None, True) else EINVAL  # noqa:E501

    def parse(self, root: Optional[add_command] = None,
              argv: Optional[Sequence[str]] = None, **kwargs) -> Namespace:
        '''Parse the command line.
        '''
        if root is None:
            root = self.root
        assert isinstance(root, add_command)

        _map: Dict[add_command, argp] = {}
        _arg = argp(argv=argv, **kwargs)
        self.__prog = _arg.prog
        self.__add_optional_version(_arg)
        # To support preparse_from_sys_argv(), all subparsers must be added
        # first. Otherwise, an error will occur during the help action.
        self.__add_parser(_map, _arg, root, **kwargs)
        self.__add_option(_map)

        args = _arg.parse_args(args=argv)
        assert isinstance(args, Namespace)
        self.__parse_logger(args)
        self.args = args
        return self.args

    def has_sub(self, root: add_command,
                args: Optional[Namespace] = None) -> bool:
        '''If the root command node has any subcommand nodes, return true.

        @param root: Command node
        @type root: add_command

        @param args: Command arguments
        @type args: Namespace or None (default self.args if None is specified)

        @return: bool
        '''
        if args is None:
            args = self.args
        assert isinstance(root, add_command)
        assert isinstance(args, Namespace)
        return isinstance(getattr(args, root.sub_dest), str)\
            if hasattr(args, root.sub_dest) else False

    def __run(self, args: Namespace, root: add_command) -> int:
        assert isinstance(root, add_command)
        assert isinstance(root.bind, run_command)

        if not root.bind.skip or not self.has_sub(root, args):
            ret = root.bind.func(self)
            if self.check_error(ret):
                return ret

        if hasattr(args, root.sub_dest):
            sub_dest = getattr(args, root.sub_dest)
            if isinstance(sub_dest, str):
                assert isinstance(root.subs, (list, tuple))
                for sub in root.subs:
                    assert isinstance(sub, add_command)
                    if sub.name == sub_dest:
                        ret = self.__run(args, sub)
                        if self.check_error(ret):
                            return ret

        done = root.bind.done
        if done is not None:
            assert isinstance(done, end_command)
            if not root.bind.skip or not self.has_sub(root, args):
                ret = done.func(self)  # purge
                if self.check_error(ret):
                    return ret
        return 0

    def __pre(self, args: Namespace, root: add_command) -> int:
        assert isinstance(root, add_command)
        assert isinstance(root.bind, run_command)

        prep = root.bind.prep
        if prep is not None:
            assert isinstance(prep, pre_command)
            if not root.bind.skip or not self.has_sub(root, args):
                ret = prep.func(self)
                if self.check_error(ret):
                    return ret

        if hasattr(args, root.sub_dest):
            sub_dest = getattr(args, root.sub_dest)
            if isinstance(sub_dest, str):
                assert isinstance(root.subs, (list, tuple))
                for sub in root.subs:
                    assert isinstance(sub, add_command)
                    if sub.name == sub_dest:
                        return self.__pre(args, sub)
        return 0

    def run(self,
            root: Optional[add_command] = None,
            argv: Optional[Sequence[str]] = None,
            **kwargs) -> int:
        '''Parse and run the command line.
        '''
        if root is None:
            root = self.root

        if not isinstance(root, add_command):
            self.logger.debug("cannot find root")
            return ENOENT

        kwargs.pop("prog", None)  # Please do not specify prog
        if "description" in root.options:  # Default use root's description
            kwargs["description"] = root.options["description"]
        args = self.parse(root, argv, **kwargs)
        self.logger.debug("%s", args)

        try:
            version = self.version
            if isinstance(version, str):
                # Output version for the debug level. Internal log
                # items are debug level only, except for errors.
                self.logger.debug("version: %s", version)

            ret = self.__pre(args, root)
            if self.check_error(ret):
                return ret
            return self.__run(args, root)
        except KeyboardInterrupt:
            return ECANCELED
        except BaseException:  # pylint: disable=broad-except
            self.logger.exception("Something went wrong:")
            return ENOTRECOVERABLE


cmds: commands = commands()



# coding:utf-8

from hashlib import md5
from hashlib import sha1
from hashlib import sha256
import os
from queue import Empty
from queue import Queue
import stat
from threading import Thread
from threading import current_thread  # noqa:H306
from typing import Callable
from typing import Dict
from typing import Generator
from typing import List
from typing import Optional
from typing import Sequence
from typing import Set

from xkits.actuator import commands

CPU_COUNT = os.cpu_count()
THDNUM_MINIMUM = 1
THDNUM_MAXIMUM = CPU_COUNT if isinstance(CPU_COUNT, int) else 64
THDNUM_DEFAULT = int(THDNUM_MAXIMUM / 2)


class scanner:
    '''scan objects
    '''

    class object:  # pylint: disable=too-many-public-methods

        def __init__(self, path: str):
            assert isinstance(path, str)
            self.__path = os.path.normpath(path)
            self.__abspath = os.path.abspath(self.__path)
            self.__realpath = os.path.realpath(self.__abspath)

        @property
        def path(self) -> str:
            return self.__path

        @property
        def abspath(self) -> str:
            return self.__abspath

        @property
        def realpath(self) -> str:
            return self.__realpath

        @property
        def stat(self) -> os.stat_result:
            return os.stat(self.abspath)

        @property
        def lstat(self) -> os.stat_result:
            return os.lstat(self.abspath)

        @property
        def uid(self) -> int:
            '''user id of owner
            '''
            return self.stat.st_uid

        @property
        def gid(self) -> int:
            '''group id of owner
            '''
            return self.stat.st_gid

        @property
        def mode(self) -> int:
            return self.stat.st_mode

        @property
        def ctime(self) -> float:
            return self.stat.st_ctime

        @property
        def atime(self) -> float:
            '''time of most recent access
            '''
            return self.stat.st_atime

        @property
        def mtime(self) -> float:
            '''time of most recent content modification
            '''
            return self.stat.st_mtime

        @property
        def size(self) -> int:
            return self.stat.st_size

        @property
        def isdir(self) -> bool:
            return stat.S_ISDIR(self.stat.st_mode)

        @property
        def isreg(self) -> bool:
            return stat.S_ISREG(self.stat.st_mode)

        @property
        def isfile(self) -> bool:
            return self.isreg

        @property
        def islink(self) -> bool:
            return stat.S_ISLNK(self.lstat.st_mode)

        @property
        def issym(self) -> bool:
            return self.islink

        def hash(self, *args, size=1024**2) -> Generator[str, None, None]:
            assert self.isfile and not self.issym
            with open(self.abspath, "rb") as fhandler:
                while True:
                    data = fhandler.read(size)
                    if not data:
                        break
                    for obj in args:
                        obj.update(data)
            return (obj.hexdigest() for obj in args)

        @property
        def md5(self) -> str:
            code, = self.hash(md5())
            return code

        @property
        def sha1(self) -> str:
            code, = self.hash(sha1())
            return code

        @property
        def sha256(self) -> str:
            code, = self.hash(sha256())
            return code

    def __init__(self):
        self.__objdict: Dict[str, scanner.object] = {}
        self.__objects: Set[scanner.object] = set()
        self.__objsyms: Set[scanner.object] = set()
        self.__objregs: Set[scanner.object] = set()
        self.__objdirs: Set[scanner.object] = set()

    def __iter__(self):
        return iter(self.__objects)

    def __getitem__(self, key: str):
        return self.__objdict[key]

    @property
    def dirs(self) -> Set[object]:
        return self.__objdirs

    @property
    def files(self) -> Set[object]:
        return self.__objregs

    @property
    def links(self) -> Set[object]:
        return self.__objsyms

    def add(self, obj: object):
        assert isinstance(obj, scanner.object)
        if obj.path not in self.__objdict:
            self.__objdict[obj.path] = obj
            self.__objects.add(obj)
            if obj.issym:
                self.__objsyms.add(obj)
            if obj.isdir:
                self.__objdirs.add(obj)
            elif obj.isreg:
                self.__objregs.add(obj)

    @classmethod
    def load(cls,  # pylint: disable=R0913,R0914,R0915,R0917
             paths: Sequence[str],
             exclude: Optional[Sequence[str]] = None,
             linkdir: bool = True,
             threads: int = THDNUM_DEFAULT,
             handler: Optional[Callable[[object], bool]] = None):
        if exclude is None:
            exclude = []

        assert isinstance(paths, Sequence)
        assert isinstance(exclude, Sequence)
        assert isinstance(linkdir, bool)
        assert isinstance(threads, int)

        cmds = commands()
        thds = min(max(THDNUM_MINIMUM, threads), THDNUM_MAXIMUM)

        def rpath(path: str) -> str:
            assert isinstance(path, str)
            return os.path.relpath(path)

        # filter files and directorys
        def path_filter() -> Set[str]:
            filter_paths: Set[str] = set()

            for path in exclude:
                filter_paths.add(rpath(path))

            return filter_paths

        class task_stat:  # pylint: disable=too-few-public-methods

            def __init__(self):
                self.exit = False
                self.handler = handler
                self.scanner = scanner()
                self.filter: Set[str] = path_filter()
                self.q_path: "Queue[str]" = Queue()
                self.q_task: "Queue[scanner.object]" = Queue(maxsize=thds * 2)

        scan_stat = task_stat()

        def task_scan_path():
            scanned_dirs = set()
            name = current_thread().name
            cmds.logger.debug("task thread[%s] start", name)
            while not scan_stat.exit or not scan_stat.q_path.empty():
                try:
                    path = scan_stat.q_path.get(timeout=0.01)
                except Empty:
                    continue

                path = rpath(path)
                assert isinstance(path, str)

                if path in scan_stat.filter or not os.path.exists(path):
                    cmds.logger.debug("scan filter %s", path)
                    scan_stat.q_path.task_done()
                    continue

                if os.path.isdir(path) and path not in scanned_dirs:
                    scanned_dirs.add(path)
                    # scan symbolic link dirs?
                    if not os.path.islink(path) or linkdir:
                        for sub in os.listdir(path=path):
                            spath = os.path.join(path, sub)
                            scan_stat.q_path.put(spath)

                ret = True
                obj = scanner.object(path)

                if isinstance(scan_stat.handler, Callable):
                    ret = scan_stat.handler(obj)
                    assert isinstance(ret, bool)

                if ret is True:
                    scan_stat.q_task.put(obj)
                scan_stat.q_path.task_done()
            cmds.logger.debug("task thread[%s] exit", name)

        def task_scan():
            name = current_thread().name
            cmds.logger.debug("task thread[%s] start", name)
            while not scan_stat.exit or not scan_stat.q_task.empty():
                try:
                    obj = scan_stat.q_task.get(timeout=0.01)
                except Empty:
                    continue

                cmds.logger.debug("scan %s", obj.path)
                scan_stat.scanner.add(obj=obj)
                scan_stat.q_task.task_done()
            cmds.logger.debug("task thread[%s] exit", name)

        task_threads: List[Thread] = []
        task_threads.append(Thread(target=task_scan, name="xkits-scan"))
        task_threads.extend([
            Thread(target=task_scan_path, name=f"xkits-scan{i}")
            for i in range(thds)
        ])

        for thread in task_threads:
            thread.start()

        for path in paths:
            scan_stat.q_path.put(path)

        scan_stat.q_path.join()
        scan_stat.q_task.join()
        scan_stat.exit = True

        for thread in task_threads:
            thread.join()

        return scan_stat.scanner



# coding=utf-8

from csv import DictReader as csv_dist_reader
from csv import DictWriter as csv_dist_writer
from csv import reader as csv_reader
from csv import writer as csv_writer
import os
from typing import Any
from typing import Callable
from typing import Dict
from typing import Generic
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple
from typing import TypeVar
from typing import Union

import openpyxl
from tabulate import TableFormat
from tabulate import tabulate as __tabulate
from wcwidth import wcswidth
import xlrd
import xlwt

from xkits.safefile import safile

FKT = TypeVar("FKT")
FVT = TypeVar("FVT")
RKT = TypeVar("RKT")
RVT = TypeVar("RVT")
CVT = TypeVar("CVT")


class cell(Generic[CVT]):
    """Cell in the custom table

    Define cells to resolve the null value issue.
    """

    def __init__(self, value: Optional[CVT] = None):
        self.value = value

    def __str__(self) -> str:
        return str(self.value if self.value is not None else "")

    @property
    def empty(self) -> bool:
        return self.value is None

    @property
    def value(self) -> Optional[CVT]:
        return self.__value

    @value.setter
    def value(self, value: Optional[CVT]):
        self.__value = value


class row(Generic[RKT, RVT]):
    """Row in the custom table
    """

    def __init__(self, values: Union[Iterable[cell[RVT]],
                                     Iterable[Optional[RVT]]]):
        self.__cells = [self.new_cell(value=value) for value in values]

    def __len__(self) -> int:
        return len(self.__cells)

    def __iter__(self) -> Iterator[cell[RVT]]:
        """all cells
        """
        return iter(self.__cells)

    def __getitem__(self, index: int) -> cell[RVT]:
        return self.__cells[index]

    def __setitem__(self, index: int,
                    value: Union[cell[RVT],
                                 Optional[RVT]]
                    ) -> None:
        self.__cells[index] = self.new_cell(value=value)

    @property
    def values(self) -> Tuple[Optional[RVT], ...]:
        """all cell values
        """
        return tuple(cell.value for cell in self)

    def append(self, value: Union[cell[RVT], Optional[RVT]]) -> None:
        self.__cells.append(self.new_cell(value))

    def extend(self, values: Union[Iterable[cell[RVT]],
                                   Iterable[Optional[RVT]]]) -> None:
        self.__cells.extend(self.new_cell(value) for value in values)

    def mapping(self, header: Tuple[RKT, ...]) -> Dict[RKT, RVT]:
        """Map the value of cells into dict
        """
        return {key: cell.value for key, cell in zip(header, self)
                if cell.value is not None}

    def new_cell(self, value: Union[cell[RVT], Optional[RVT]]) -> cell[RVT]:
        return value if isinstance(value, cell) else cell(value)


class form(Generic[FKT, FVT]):
    """Custom table
    """

    def __init__(self, name: str, header: Optional[Iterable[FKT]] = None):
        self.__rows: List[row[FKT, FVT]] = []
        self.__name: str = name
        self.header = header if header is not None else []

    def __len__(self) -> int:
        return len(self.__rows)

    def __iter__(self) -> Iterator[row[FKT, FVT]]:
        """all rows
        """
        return iter(self.__rows)

    def __getitem__(self, index: int) -> row[FKT, FVT]:
        return self.__rows[index]

    def __setitem__(self, index: int,
                    value: Union[row[FKT, FVT],
                                 Iterable[cell[FVT]],
                                 Iterable[FVT]]
                    ) -> None:
        self.__rows[index] = self.new_row(value)

    @property
    def name(self) -> str:
        """table name
        """
        return self.__name

    @property
    def header(self) -> Tuple[FKT, ...]:
        """table header (title line)
        """
        return self.__header

    @header.setter
    def header(self, value: Iterable[FKT]) -> None:
        self.__header: Tuple[FKT, ...] = tuple(i for i in value)

    @property
    def mappings(self) -> Iterator[Dict[FKT, FVT]]:
        return iter(row.mapping(self.header) for row in self)

    @property
    def values(self) -> Tuple[Tuple[Optional[FVT], ...], ...]:
        """all cell values (by row)
        """
        return tuple(row.values for row in self)

    def column_no(self, key: FKT) -> int:
        return self.header.index(key)

    def sort(self, fn: Callable[[row[FKT, FVT]], cell[FVT]],
             reverse: bool = False) -> None:
        """sort rows using a Lambda function as the key.
        """
        self.__rows.sort(key=lambda row: fn(row).value,  # type: ignore
                         reverse=reverse)

    def dump(self) -> Tuple[Tuple[Any, ...], ...]:
        """dump header and all rows
        """
        table: List[Tuple[Any, ...]] = [self.header]
        table.extend(self.values)
        return tuple(table)

    def reflection(self, cells: Dict[FKT, FVT],
                   default: Any = None) -> row[FKT, FVT]:
        """Re-map the dict to new row object
        """
        return self.new_row(cells=tuple(cells.get(key, default)
                                        for key in self.header))

    def append(self, item: Union[row[FKT, FVT],
                                 Iterable[cell[FVT]],
                                 Iterable[FVT]]
               ) -> None:
        self.__rows.append(self.new_row(item))

    def extend(self, rows: Iterable[Union[row[FKT, FVT],
                                          Iterable[cell[FVT]],
                                          Iterable[FVT]]]) -> None:
        self.__rows.extend(self.new_row(row) for row in rows)

    def new_row(self, cells: Union[row[FKT, FVT],
                                   Iterable[cell[FVT]],
                                   Iterable[Optional[FVT]]]
                ) -> row[FKT, FVT]:
        return cells if isinstance(cells, row) else row(values=cells)

    def new_map(self, default: Any = None) -> Dict[FKT, Any]:
        """Generate new mapping with default values
        """
        return {key: default for key in self.header}


def tabulate(table: form[Any, Any],
             fmt: Union[str, TableFormat] = "simple") -> str:
    return __tabulate(tabular_data=table.values,
                      headers=table.header,
                      tablefmt=fmt)


def parse_table_name(filename: str) -> str:
    return os.path.splitext(os.path.basename(filename))[0]


class csv():

    @classmethod
    def load(cls, filename: str,
             include_header: bool = True
             ) -> form[str, str]:
        """Read .csv file
        """
        with safile.lock(filename):
            safile.restore(path=filename)
            table: form[str, str] = form(name=parse_table_name(filename))
            with open(filename, "r", encoding="utf-8") as rhdl:
                if include_header:
                    reader = csv_dist_reader(rhdl)
                    fields = reader.fieldnames
                    if fields is not None:
                        table.header = fields
                        for _row in reader:
                            table.append(table.reflection(_row))
                else:
                    reader = csv_reader(rhdl)
                    for _row in reader:
                        table.append(_row)
            return table

    @classmethod
    def dump(cls, filename: str, table: form[Any, Any]) -> None:
        """Write .csv file
        """
        with safile.lock(filename):
            safile.create_backup(path=filename, copy=True)
            with open(filename, "w", encoding="utf-8") as whdl:
                if len(table.header) > 0:
                    writer = csv_dist_writer(whdl, fieldnames=table.header)
                    writer.writeheader()
                    writer.writerows(table.mappings)
                else:
                    writer = csv_writer(whdl)
                    writer.writerows(table.dump())
            safile.delete_backup(path=filename)


class xls_reader():
    """Read .xls file
    """

    def __init__(self, filename: str):
        self.__file: str = filename
        safile.restore(path=filename)
        self.__book: xlrd.Book = xlrd.open_workbook(filename)

    @property
    def file(self) -> str:
        return self.__file

    @property
    def book(self) -> xlrd.Book:
        return self.__book

    def load_sheet(self, sheet_name: Optional[str] = None) -> form[str, str]:
        sheet_index: int = self.book.sheet_names().index(sheet_name)\
            if isinstance(sheet_name, str) else 0
        sheet: xlrd.sheet.Sheet = self.book.sheet_by_index(sheet_index)
        first: Iterable[str] = sheet.row_values(0)  # first line as header
        table: form[str, Any] = form(name=sheet.name, header=first)
        for i in range(1, sheet.nrows):
            table.append(sheet.row_values(i))
        return table

    def load_sheets(self) -> Tuple[form[str, str], ...]:
        return tuple(self.load_sheet(name) for name in self.book.sheet_names())


class xls_writer():
    """Write .xls file
    """
    WIDTH = 325

    def __init__(self):
        self.__book: xlwt.Workbook = xlwt.Workbook(
            encoding="utf-8", style_compression=0)

    @property
    def book(self) -> xlwt.Workbook:
        return self.__book

    def save(self, filename: str) -> bool:
        abspath: str = os.path.abspath(filename)
        try:
            dirname: str = os.path.dirname(abspath)
            if not os.path.exists(dirname):
                os.makedirs(dirname)
            safile.create_backup(path=abspath, copy=True)
            self.book.save(abspath)
            safile.delete_backup(path=abspath)
            return True
        except Exception:  # pragma: no cover, pylint: disable=broad-except
            # f"failed to write file {abspath}"
            return False  # pragma: no cover

    def dump_sheet(self, table: form[Any, Any]):
        sheet: xlwt.Worksheet = self.book.add_sheet(
            table.name, cell_overwrite_ok=True)
        widths: List[int] = []
        values: Tuple[Tuple[Any, ...], ...] = table.dump()
        for row_no, cells in enumerate(values):
            for col_no, _cell in enumerate(cells):
                value = str(_cell)
                sheet.write(row_no, col_no, value)
                width = wcswidth(value)
                if col_no >= len(widths):
                    sheet.col(col_no).width = self.WIDTH
                    widths.append(0)
                if width > widths[col_no]:
                    sheet.col(col_no).width = self.WIDTH * width
                    widths[col_no] = width

    def dump_sheets(self, tables: Iterable[form[Any, Any]]):
        for table in tables:
            self.dump_sheet(table=table)


class xlsx():
    """Read or write .xlsx file
    """

    def __init__(self, filename: str, read_only: bool = True):
        self.__file: str = filename
        safile.restore(path=filename)
        self.__book: openpyxl.Workbook = openpyxl.load_workbook(
            filename=filename, read_only=read_only)

    @property
    def file(self) -> str:
        return self.__file

    @property
    def book(self) -> openpyxl.Workbook:
        return self.__book

    def load_sheet(self, sheet_name: Optional[str] = None) -> form[str, Any]:
        def get_default_sheet_name() -> str:
            if isinstance(sheet_name, str):
                return sheet_name
            active = self.book.active
            return self.book.sheetnames[0] if active is None else active.title

        sheet = self.book[get_default_sheet_name()]
        first = list(sheet.iter_rows(max_row=1))[0]
        cells: List[str] = [c.value for c in first if isinstance(c.value, str)]
        table: form[str, Any] = form(name=sheet.title, header=cells)
        for _row in sheet.iter_rows(min_row=2):
            table.append([cell.value for cell in _row])
        return table

    def load_sheets(self) -> Tuple[form[str, str], ...]:
        return tuple(self.load_sheet(name) for name in self.book.sheetnames)



# coding:utf-8

from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import TimeoutError as ThreadTimeout
from typing import Callable
from typing import Union

ExecuteTimeUnit = Union[float, int]


class Executor():  # pylint: disable=too-few-public-methods
    def __init__(self, fn: Callable, *args, **kwargs) -> None:
        self.__fn = fn
        self.__args = args
        self.__kwargs = kwargs

    def countdown(self, seconds: ExecuteTimeUnit):
        with ThreadPoolExecutor() as executor:
            try:
                future = executor.submit(self.__fn, *self.__args, **self.__kwargs)  # noqa:E501
                return future.result(seconds)
            except ThreadTimeout as exc:
                message: str = f"Run timeout of {seconds} seconds"
                raise TimeoutError(message) from exc


def hourglass(seconds: ExecuteTimeUnit):
    def decorator(fn):
        def inner(*args, **kwargs):
            return Executor(fn, *args, **kwargs).countdown(seconds)
        return inner
    return decorator



# coding:utf-8

from urllib.parse import urljoin

__project__ = "xkits"
__version__ = "2.4.0"
__description__ = "Python toolkit."
__urlhome__ = f"https://github.com/bondbox/{__project__}/"
__urlcode__ = __urlhome__
__urldocs__ = __urlhome__
__urlbugs__ = urljoin(__urlhome__, "issues")

# author
__author__ = "Mingzhe Zou"
__author_email__ = "zoumingzhe@outlook.com"

# project name
__prog_name__ = "xkits"
__prog_complete__ = "xargcomplete"
__prog_project__ = "xargproject"



# coding:utf-8

from xkits.actuator import Namespace  # noqa:F401
from xkits.actuator import add_command  # noqa:F401,H306
from xkits.actuator import cmds  # noqa:F401
from xkits.actuator import commands  # noqa:F401
from xkits.actuator import end_command  # noqa:F401
from xkits.actuator import pre_command  # noqa:F401
from xkits.actuator import run_command  # noqa:F401
from xkits.cache import CacheAtom  # noqa:F401
from xkits.cache import CacheData  # noqa:F401
from xkits.cache import CacheExpired  # noqa:F401
from xkits.cache import CacheItem  # noqa:F401
from xkits.cache import CacheLookupError  # noqa:F401
from xkits.cache import CacheMiss  # noqa:F401
from xkits.cache import CachePool  # noqa:F401
from xkits.cache import CacheTimeUnit  # noqa:F401
from xkits.cache import ItemPool  # noqa:F401
from xkits.cache import NamedCache  # noqa:F401
from xkits.colorful import Back  # noqa:F401
from xkits.colorful import Fore  # noqa:F401
from xkits.colorful import Style  # noqa:F401
from xkits.colorful import color  # noqa:F401,H306
from xkits.executor import hourglass  # noqa:F401
from xkits.parser import argp  # noqa:F401
from xkits.safefile import safile  # noqa:F401
from xkits.safefile import stfile  # noqa:F401
from xkits.scanner import scanner  # noqa:F401
from xkits.sheet import cell  # noqa:F401
from xkits.sheet import csv  # noqa:F401
from xkits.sheet import form  # noqa:F401
from xkits.sheet import row  # noqa:F401
from xkits.sheet import tabulate  # noqa:F401
from xkits.sheet import xls_reader  # noqa:F401
from xkits.sheet import xls_writer  # noqa:F401
from xkits.sheet import xlsx  # noqa:F401
from xkits.sitepage import Page  # noqa:F401
from xkits.sitepage import PageCache  # noqa:F401
from xkits.sitepage import ProxyProtocol  # noqa:F401
from xkits.sitepage import ProxySession  # noqa:F401
from xkits.sitepage import Site  # noqa:F401
from xkits.thread import NamedLock  # noqa:F401
from xkits.thread import TaskJob  # noqa:F401
from xkits.thread import TaskPool  # noqa:F401
from xkits.thread import ThreadPool  # noqa:F401
from xkits.utils import chdir  # noqa:F401
from xkits.utils import singleton  # noqa:F401



# coding:utf-8

from argparse import ArgumentParser
from argparse import Namespace
from argparse import _ArgumentGroup  # noqa:H306
from argparse import _HelpAction
from argparse import _SubParsersAction
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Set
from typing import Tuple

from xkits.attribute import __project__
from xkits.attribute import __urlhome__

try:
    from argcomplete import autocomplete
except ModuleNotFoundError:  # pragma: no cover
    pass  # pragma: no cover


class checker():

    prefix_chars = "-"

    @classmethod
    def check_name_pos(cls, fn):
        '''check positional argument name
        '''

        def inner(self, name: str, **kwargs):
            assert isinstance(name, str) and name[0] not in cls.prefix_chars, \
                f"{name} is not a positional argument name"
            return fn(self, name, **kwargs)

        return inner

    @classmethod
    def check_name_opt(cls, fn):
        '''check optional argument name
        '''

        def inner(self, *name: str, **kwargs):
            # 1. check short form optional argument ("-x")
            # 2. check long form optional argument ("--xx")
            # 3. only short form or long form or short form + long form
            # 模棱两可的数据（-1可以是一个负数的位置参数）
            for n in name:
                assert isinstance(n, str) and n[0] in cls.prefix_chars, \
                    f"{n} is not an optional argument name"
            return fn(self, *name, **kwargs)

        return inner

    @classmethod
    def check_nargs_opt(cls, fn):
        '''nargs hook function:
            nargs < -1: using "?", 0 or 1 argument, default value
            nargs = -1: using "+", arguments list, at least 1
            nargs = 0: using "*", arguments list, allow to be empty
            nargs = 1: redirect to "?", 0 or 1 argument
            nargs > 1: N arguments list
        '''

        def inner(self, *args, **kwargs):
            _nargs = kwargs.get("nargs", -2)
            if isinstance(_nargs, int) and _nargs < 2:
                _nargs = {1: "?", 0: "*", -1: "+"}.get(_nargs, "?")
            kwargs.update({"nargs": _nargs})
            return fn(self, *args, **kwargs)

        return inner


class argp(ArgumentParser):
    '''Simple command-line tool based on argparse.
    '''

    def __init__(self,  # pylint: disable=R0913,R0917
                 argv: Optional[Sequence[str]] = None,
                 prog: Optional[str] = None,
                 usage: Optional[str] = None,
                 prev_parser: Optional["argp"] = None,
                 description: Optional[str] = f"Command-line based on {__project__}.",  # noqa:E501
                 epilog: Optional[str] = f"For more, please visit {__urlhome__}",  # noqa:E501
                 **kwargs):
        kwargs.setdefault("prog", prog)
        kwargs.setdefault("usage", usage)
        kwargs.setdefault("description", description)
        kwargs.setdefault("epilog", epilog)
        ArgumentParser.__init__(self, **kwargs)
        self.__argv: Optional[Sequence[str]] = argv
        self.__help_option: Dict[str, _HelpAction] = {}
        self.__prev_parser: argp = prev_parser or self
        self.__next_parser: List[argp] = []
        if prev_parser is not None:
            prev_parser.next_parser.append(self)

    @property
    def argv(self) -> Optional[Sequence[str]]:
        root = self.root_parser
        return root.argv if root is not self else self.__argv

    @property
    def prev_parser(self) -> "argp":
        return self.__prev_parser

    @property
    def next_parser(self) -> List["argp"]:
        return self.__next_parser

    @property
    def root_parser(self) -> "argp":
        root = self.prev_parser
        while root.prev_parser != root:
            root = root.prev_parser
        return root

    def argument_group(self,
                       title: Optional[str] = None,
                       description: Optional[str] = None,
                       **kwargs) -> _ArgumentGroup:
        '''Find the created argument group by title, create if not exist.
        '''
        for group in self._action_groups:
            if title == group.title:
                return group
        return self.add_argument_group(title, description, **kwargs)

    @checker.check_name_opt
    def filter_optional_name(self, *name: str) -> Sequence[str]:
        '''Filter defined optional argument name.
        '''
        option_strings: Set[str] = set()
        for action in self._get_optional_actions():
            option_strings.update(action.option_strings)
        return [n for n in name if n not in option_strings]

    @checker.check_name_pos
    def add_pos(self, name: str, **kwargs) -> None:
        '''Add positional argument.
        '''
        assert "dest" not in kwargs, \
            "dest supplied twice for positional argument"
        self.add_argument(name, **kwargs)

    @checker.check_name_opt
    @checker.check_nargs_opt
    def add_opt(self, *name: str, **kwargs) -> None:
        '''Add optional argument.
        '''
        self.add_argument(*name, **kwargs)

    @checker.check_name_opt
    def add_opt_on(self, *name: str, **kwargs) -> None:
        '''Add boolean optional argument, default value is False.
        '''
        kwargs.update({"action": "store_true"})
        for key in ("type", "nargs", "const", "default", "choices"):
            assert key not in kwargs, f"'{key}' is an invalid argument"
        self.add_argument(*name, **kwargs)

    @checker.check_name_opt
    def add_opt_off(self, *name: str, **kwargs) -> None:
        '''Add boolean optional argument, default value is True.
        '''
        kwargs.update({"action": "store_false"})
        for key in ("type", "nargs", "const", "default", "choices"):
            assert key not in kwargs, f"'{key}' is an invalid argument"
        self.add_argument(*name, **kwargs)

    def add_subparsers(self, *args, **kwargs) -> _SubParsersAction:
        '''Add subparsers.
        '''
        # subparser: cannot have multiple subparser arguments
        kwargs.setdefault("title", "subcommands")
        kwargs.setdefault("description", None)
        kwargs.setdefault("dest", f"subcmd_{self.prog}")
        kwargs.setdefault("help", None)
        kwargs.setdefault("metavar", None)
        return ArgumentParser.add_subparsers(self, *args, **kwargs)

    def parse_args(  # pylint: disable=useless-parent-delegation
        self, args: Optional[Sequence[str]] = None,
        namespace: Optional[Namespace] = None
    ) -> Namespace:
        try:
            autocomplete(self)  # For tab completion
        except NameError:  # pragma: no cover
            pass  # pragma: no cover
        return super().parse_args(args, namespace)  # type: ignore

    def parse_known_args(  # pylint: disable=useless-parent-delegation
            self, args: Optional[Sequence[str]] = None,
            namespace: Optional[Namespace] = None
    ) -> Tuple[Namespace, List[str]]:
        return super().parse_known_args(args, namespace)  # type: ignore

    def __enable_help_action(self):  # pylint: disable=unused-private-member
        while len(self.__help_option) > 0:
            option, action = self.__help_option.popitem()
            self._option_string_actions[option] = action
        assert len(self.__help_option) == 0

    def __disable_help_action(self):  # pylint: disable=unused-private-member
        assert len(self.__help_option) == 0
        for option, action in self._option_string_actions.items():
            if isinstance(action, _HelpAction):
                self.__help_option[option] = action
        for option in self.__help_option:
            self._option_string_actions.pop(option)

    def preparse_from_sys_argv(self) -> Namespace:
        '''Preparse some arguments from sys.argv for tab completion.

        When arguments contain the help option, call parse_known_args()
        will print help message and exit. The command line can parse
        normally.

        But parameters added after calling preparse_from_sys_argv() will
        not show in the help message, because the exit occurred before
        adding parameters.

        So, disable the help action before calling parse_known_args().
        The help option will be stored, and restored after the call ends.
        '''

        def __dfs_enable_help_action(root: argp):
            root.__enable_help_action()  # pylint: disable=protected-access
            for _sub in root.next_parser:
                __dfs_enable_help_action(_sub)

        def __dfs_disable_help_action(root: argp):
            root.__disable_help_action()  # pylint: disable=protected-access
            for _sub in root.next_parser:
                __dfs_disable_help_action(_sub)

        try:
            __dfs_disable_help_action(self.root_parser)
            namespace, _ = self.root_parser.parse_known_args(self.argv)
            return namespace
        finally:
            __dfs_enable_help_action(self.root_parser)



# coding:utf-8

from base64 import b16decode
from base64 import b16encode
from configparser import ConfigParser
from errno import EPERM
import os
import shutil
import sys
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Sequence
from typing import Set
from typing import Tuple
from typing import Union

from pip._internal.commands.show import _PackageInfo
from pip._internal.commands.show import search_packages_info
from tabulate import tabulate

from xkits.actuator import add_command
from xkits.actuator import commands
from xkits.actuator import run_command
from xkits.attribute import __prog_complete__
from xkits.attribute import __project__
from xkits.attribute import __urlhome__
from xkits.attribute import __version__
from xkits.parser import argp
from xkits.utils import singleton

USER_BASH_COMPLETION_CFG = "~/.bash_completion"
USER_BASH_COMPLETION_DIR = "~/.bash_completion.d"


def enable_bash():

    def update_code():
        bash_completion_path = os.path.expanduser(USER_BASH_COMPLETION_CFG)
        bash_completion_hook = os.path.expanduser(USER_BASH_COMPLETION_DIR)
        bash_completion_code = """
for bcfile in ~/.bash_completion.d/* ; do
  source ${bcfile}
done
"""
        if not os.path.exists(bash_completion_hook):
            os.makedirs(bash_completion_hook)
        with open(bash_completion_path, "r", encoding="utf-8") as fh:
            if bash_completion_code in fh.read():
                return
        with open(bash_completion_path, "a", encoding="utf-8") as fh:
            fh.write(f"\n{bash_completion_code}\n")

    update_code()
    update_bash(__prog_complete__)


def update_bash(cmd: str) -> int:
    bash_completion_hook = os.path.expanduser(USER_BASH_COMPLETION_DIR)
    name = b16encode(cmd.encode()).decode()
    path = os.path.join(bash_completion_hook, f"{__prog_complete__}-{name}")
    return os.system(f"register-python-argcomplete {cmd} > {path}")


def remove_bash(cmd: str) -> bool:
    bash_completion_hook = os.path.expanduser(USER_BASH_COMPLETION_DIR)
    name = b16encode(cmd.encode()).decode()
    path = os.path.join(bash_completion_hook, f"{__prog_complete__}-{name}")
    if os.path.isfile(path):
        os.remove(path)
    return not os.path.exists(path)


def list_bash() -> Set[str]:
    cmds: Set[str] = set()
    bash_completion_hook = os.path.expanduser(USER_BASH_COMPLETION_DIR)
    if not os.path.exists(bash_completion_hook):
        os.makedirs(bash_completion_hook)
    for item in os.listdir(bash_completion_hook):
        if not os.path.isfile(os.path.join(bash_completion_hook, item)):
            continue
        keys = item.split("-", 1)
        if len(keys) == 2 and keys[0] == __prog_complete__:
            cmds.add(b16decode(keys[1]).decode())
    return cmds


@singleton
class collections:

    def __init__(self):
        self.__cmds: Set[str] = set()
        for _pkg in tuple({"argcomplete", __project__}):
            for _req in set(self.get_package_info(_pkg).required_by):
                config = ConfigParser()
                package_info = self.get_package_info(_req)
                config.read_string(os.linesep.join(package_info.entry_points))
                if config.has_section("console_scripts"):
                    for _cmd in config["console_scripts"]:
                        self.__cmds.add(_cmd)

    @property
    def cmds(self) -> Iterable[str]:
        return iter(self.__cmds)

    @classmethod
    def get_package_info(cls, package_name: str) -> _PackageInfo:
        return list(search_packages_info([package_name]))[0]


@add_command("enable", help="Enable completion.")
def add_cmd_enable(_arg: argp):
    pass


@run_command(add_cmd_enable)
def run_cmd_enable(cmds: commands) -> int:
    which = shutil.which("activate-global-python-argcomplete")
    command = f"{which} --user --yes"
    cmds.logger.info(command)
    retcode = os.system(command)
    if retcode != 0:
        return retcode
    enable_bash()
    return 0


@add_command("update", help="Update completion config.")
def add_cmd_update(_arg: argp):
    group_script = _arg.argument_group("Specify update command or script")
    group_script.add_argument("--script", type=str, nargs=1, metavar="PATH",
                              dest="_commands_", action="extend",
                              help="Specify script path.")
    group_script.add_argument(type=str, nargs="*", metavar="command",
                              dest="_commands_", action="extend",
                              help="Specify command.")


@run_command(add_cmd_update)
def run_cmd_update(cmds: commands) -> int:
    iter_commands = getattr(cmds.args, "_commands_")
    if len(iter_commands) == 0:
        iter_commands = collections().cmds
    for cmd in set(iter_commands):
        if shutil.which(cmd) is None:
            cmds.stderr(f"Non existent command or script: {cmd}")
            continue
        cmds.stdout(f"Update command or script: {cmd}")
        update_bash(cmd)
    cmds.stdout("Please restart your shell or source the file to activate it.")
    cmds.stdout(f"Bash: source {os.path.expanduser(USER_BASH_COMPLETION_CFG)}")
    return 0


@add_command("remove", help="Remove completion config.")
def add_cmd_remove(_arg: argp):
    allowed = list(list_bash())
    group_script = _arg.argument_group("Specify remove command or script")
    mgroup_script = group_script.add_mutually_exclusive_group()
    mgroup_script.add_argument("--auto-clean", dest="_clean_",
                               action="store_true",
                               help="Clean invalid commands or scripts.")
    mgroup_script.add_argument("--all", const=allowed,
                               dest="_commands_", action="store_const",
                               help="Remove all commands or scripts.")
    group_script.add_argument(type=str, nargs="*", metavar="command",
                              dest="_commands_", action="extend",
                              help="Specify command or script.",
                              choices=allowed + [[]])


@run_command(add_cmd_remove)
def run_cmd_remove(cmds: commands) -> int:
    list_commands: List[str] = getattr(cmds.args, "_commands_")
    if getattr(cmds.args, "_clean_"):
        assert isinstance(list_commands, list)
        for cmd in list_bash():
            if cmd in list_commands:
                continue
            if shutil.which(cmd) is None:
                list_commands.append(cmd)
    for cmd in set(list_commands):
        cmds.stdout(f"Remove command or script: {cmd}")
        assert remove_bash(cmd)
    return 0


@add_command("list", help="List all completion.")
def add_cmd_list(_arg: argp):
    pass


@run_command(add_cmd_list)
def run_cmd_list(cmds: commands) -> int:
    table: Dict[str, Dict[str, Union[None, str, Set[str]]]] = {}

    def update_table(cmd: str, shell: str):
        if cmd not in table:
            table[cmd] = {"which": shutil.which(cmd), "shell": set()}
        shell_set = table[cmd]["shell"]
        assert isinstance(shell_set, set)
        shell_set.add(shell)

    def output_table() -> List[Tuple[str, str, str]]:
        datas: List[Tuple[str, str, str]] = []
        for k, v in table.items():
            which = v["which"] if v["which"] is not None else "None"
            shell = v["shell"]
            assert isinstance(which, str)
            assert isinstance(shell, set)
            datas.append((k, which, ", ".join(shell)))
        datas.sort(key=lambda line: line[0])
        datas.insert(0, ("command", "which", "shell"))
        return datas

    for cmd in list_bash():
        update_table(cmd, "bash")
    cmds.stdout(tabulate(output_table(), headers="firstrow"))
    return 0


@add_command(__prog_complete__, description="Tab completion management.")
def add_cmd(_arg: argp):
    pass


@run_command(add_cmd, add_cmd_enable, add_cmd_update, add_cmd_remove,
             add_cmd_list)
def run_cmd(cmds: commands) -> int:
    if sys.version_info < (3, 8):
        cmds.logger.error("Require Python>=3.8")
        return EPERM
    return 0


def main(argv: Optional[Sequence[str]] = None) -> int:
    cmds = commands()
    cmds.version = __version__
    return cmds.run(root=add_cmd, argv=argv, epilog=f"For more, please visit {__urlhome__}.")  # noqa:E501



# coding:utf-8

from typing import Optional
from typing import Set
from typing import Union

from colorama import Back
from colorama import Fore
from colorama.ansi import AnsiStyle
from colorama.ansi import code_to_chars


class AnsiXStyle(AnsiStyle):  # pylint: disable=too-few-public-methods
    ITALIC = 3
    UNDERLINE = 4
    SLOWBLINK = 5
    RAPIDBLINK = 6
    INVERT = 7
    HIDE = 8
    STRIKETHROUGH = 9
    DOUBLYUNDERLINED = 21
    RESET_ITALIC = 23
    RESET_UNDERLINE = 24
    RESET_BLINK = 25
    RESET_INVERT = 27
    REVEAL = 28
    RESET_STRIKETHROUGH = 29


Style = AnsiXStyle()
StyleType = Union[str, int]
SytleReset = {
    Style.BRIGHT: Style.NORMAL,
    Style.DIM: Style.NORMAL,
    Style.ITALIC: Style.RESET_ITALIC,
    Style.UNDERLINE: Style.RESET_UNDERLINE,
    Style.SLOWBLINK: Style.RESET_BLINK,
    Style.RAPIDBLINK: Style.RESET_BLINK,
    Style.INVERT: Style.RESET_INVERT,
    Style.STRIKETHROUGH: Style.RESET_STRIKETHROUGH,
    Style.DOUBLYUNDERLINED: Style.RESET_UNDERLINE,
}


class color(str):  # pylint: disable=too-many-public-methods
    '''Colorful terminal text

    Reference: https://en.wikipedia.org/wiki/ANSI_escape_code
    '''

    def __init__(self, obj: object):  # pylint: disable=unused-argument
        self.__background: Optional[str] = None
        self.__foreground: Optional[str] = None
        self.__style: Set[str] = set()
        super().__init__()

    def __str__(self) -> str:
        message: str = super().__str__()

        if isinstance(self.foreground, str):
            message = f"{self.foreground}{message}{Fore.RESET}"

        if isinstance(self.background, str):
            message = f"{self.background}{message}{Back.RESET}"

        if len(self.style) > 0:
            for style in self.style:
                reset_style = SytleReset.get(style, Style.RESET_ALL)
                message = f"{style}{message}{reset_style}"
        return message

    @property
    def background(self) -> Optional[str]:
        return self.__background

    @background.setter
    def background(self, value: str):
        assert isinstance(value, str), f"Unexpected type: {type(value)}"
        self.__background = value

    @property
    def foreground(self) -> Optional[str]:
        return self.__foreground

    @foreground.setter
    def foreground(self, value: str):
        assert isinstance(value, str), f"Unexpected type: {type(value)}"
        self.__foreground = value

    @property
    def style(self) -> Set[str]:
        return self.__style

    @style.setter
    def style(self, value: Set[StyleType]):
        assert isinstance(value, set), f"Unexpected type: {type(value)}"
        _value: Set[str] = {code_to_chars(v) if isinstance(v, int) else v
                            for v in value}
        self.__style = _value

    def add_style(self, value: StyleType) -> "color":
        assert isinstance(value, (str, int)), f"Unexpected type: {type(value)}"
        _value: str = code_to_chars(value) if isinstance(value, int) else value
        self.__style.add(_value)
        return self

    @classmethod
    def new_background(cls, obj: object, value: str) -> "color":
        colour: color = color(obj)
        colour.background = value
        return colour

    @classmethod
    def new_foreground(cls, obj: object, value: str) -> "color":
        colour: color = color(obj)
        colour.foreground = value
        return colour

    @classmethod
    def new_style(cls, obj: object, value: Set[StyleType]) -> "color":
        colour: color = color(obj)
        colour.style = value
        return colour

    @classmethod
    def reset(cls) -> str:
        return Style.RESET_ALL

    @classmethod
    def bold(cls, obj: object) -> "color":
        return color.new_style(obj, {Style.BRIGHT})

    @classmethod
    def dim(cls, obj: object) -> "color":
        return color.new_style(obj, {Style.DIM})

    @classmethod
    def italic(cls, obj: object) -> "color":
        return color.new_style(obj, {Style.ITALIC})

    @classmethod
    def underline(cls, obj: object) -> "color":
        return color.new_style(obj, {Style.UNDERLINE})

    @classmethod
    def slow_blink(cls, obj: object) -> "color":
        return color.new_style(obj, {Style.SLOWBLINK})

    @classmethod
    def rapid_blink(cls, obj: object) -> "color":
        return color.new_style(obj, {Style.RAPIDBLINK})

    @classmethod
    def invert(cls, obj: object) -> "color":
        return color.new_style(obj, {Style.INVERT})

    @classmethod
    def hide(cls, obj: object) -> "color":
        return color.new_style(obj, {Style.HIDE})

    @classmethod
    def strikethrough(cls, obj: object) -> "color":
        return color.new_style(obj, {Style.STRIKETHROUGH})

    @classmethod
    def doubly_underlined(cls, obj: object) -> "color":
        return color.new_style(obj, {Style.DOUBLYUNDERLINED})

    @classmethod
    def normal(cls, obj: object) -> "color":
        return color.new_style(obj, {Style.NORMAL})

    @classmethod
    def reveal(cls, obj: object) -> "color":
        return color.new_style(obj, {Style.REVEAL})

    @classmethod
    def black(cls, obj: object) -> "color":
        return color.new_foreground(obj, Fore.BLACK)

    @classmethod
    def red(cls, obj: object) -> "color":
        return color.new_foreground(obj, Fore.RED)

    @classmethod
    def green(cls, obj: object) -> "color":
        return color.new_foreground(obj, Fore.GREEN)

    @classmethod
    def yellow(cls, obj: object) -> "color":
        return color.new_foreground(obj, Fore.YELLOW)

    @classmethod
    def blue(cls, obj: object) -> "color":
        return color.new_foreground(obj, Fore.BLUE)

    @classmethod
    def magenta(cls, obj: object) -> "color":
        return color.new_foreground(obj, Fore.MAGENTA)

    @classmethod
    def cyan(cls, obj: object) -> "color":
        return color.new_foreground(obj, Fore.CYAN)

    @classmethod
    def white(cls, obj: object) -> "color":
        return color.new_foreground(obj, Fore.WHITE)

    @classmethod
    def lightblack(cls, obj: object) -> "color":
        return color.new_foreground(obj, Fore.LIGHTBLACK_EX)

    @classmethod
    def lightred(cls, obj: object) -> "color":
        return color.new_foreground(obj, Fore.LIGHTRED_EX)

    @classmethod
    def lightgreen(cls, obj: object) -> "color":
        return color.new_foreground(obj, Fore.LIGHTGREEN_EX)

    @classmethod
    def lightyellow(cls, obj: object) -> "color":
        return color.new_foreground(obj, Fore.LIGHTYELLOW_EX)

    @classmethod
    def lightblue(cls, obj: object) -> "color":
        return color.new_foreground(obj, Fore.LIGHTBLUE_EX)

    @classmethod
    def lightmagenta(cls, obj: object) -> "color":
        return color.new_foreground(obj, Fore.LIGHTMAGENTA_EX)

    @classmethod
    def lightcyan(cls, obj: object) -> "color":
        return color.new_foreground(obj, Fore.LIGHTCYAN_EX)

    @classmethod
    def lightwhite(cls, obj: object) -> "color":
        return color.new_foreground(obj, Fore.LIGHTWHITE_EX)

    @classmethod
    def black_back(cls, obj: object) -> "color":
        return color.new_background(obj, Back.BLACK)

    @classmethod
    def red_back(cls, obj: object) -> "color":
        return color.new_background(obj, Back.RED)

    @classmethod
    def green_back(cls, obj: object) -> "color":
        return color.new_background(obj, Back.GREEN)

    @classmethod
    def yellow_back(cls, obj: object) -> "color":
        return color.new_background(obj, Back.YELLOW)

    @classmethod
    def blue_back(cls, obj: object) -> "color":
        return color.new_background(obj, Back.BLUE)

    @classmethod
    def magenta_back(cls, obj: object) -> "color":
        return color.new_background(obj, Back.MAGENTA)

    @classmethod
    def cyan_back(cls, obj: object) -> "color":
        return color.new_background(obj, Back.CYAN)

    @classmethod
    def white_back(cls, obj: object) -> "color":
        return color.new_background(obj, Back.WHITE)

    @classmethod
    def lightblack_back(cls, obj: object) -> "color":
        return color.new_background(obj, Back.LIGHTBLACK_EX)

    @classmethod
    def lightred_back(cls, obj: object) -> "color":
        return color.new_background(obj, Back.LIGHTRED_EX)

    @classmethod
    def lightgreen_back(cls, obj: object) -> "color":
        return color.new_background(obj, Back.LIGHTGREEN_EX)

    @classmethod
    def lightyellow_back(cls, obj: object) -> "color":
        return color.new_background(obj, Back.LIGHTYELLOW_EX)

    @classmethod
    def lightblue_back(cls, obj: object) -> "color":
        return color.new_background(obj, Back.LIGHTBLUE_EX)

    @classmethod
    def lightmagenta_back(cls, obj: object) -> "color":
        return color.new_background(obj, Back.LIGHTMAGENTA_EX)

    @classmethod
    def lightcyan_back(cls, obj: object) -> "color":
        return color.new_background(obj, Back.LIGHTCYAN_EX)

    @classmethod
    def lightwhite_back(cls, obj: object) -> "color":
        return color.new_background(obj, Back.LIGHTWHITE_EX)



# coding:utf-8

from threading import Lock
from time import time
from typing import Any
from typing import Dict
from typing import Generic
from typing import Iterator
from typing import Optional
from typing import TypeVar
from typing import Union

CacheTimeUnit = Union[float, int]


class CacheLookupError(LookupError):
    pass


class CacheMiss(CacheLookupError):
    def __init__(self, name: Any):
        super().__init__(f"Not found {name} in cache")


class CacheExpired(CacheLookupError):
    def __init__(self, name: Optional[Any] = None):
        super().__init__("Cache expired" if name is None else f"Cache {name} expired")  # noqa:E501


CADT = TypeVar("CADT")


class CacheAtom(Generic[CADT]):
    '''Data cache without name'''

    def __init__(self, data: CADT, lifetime: CacheTimeUnit = 0):
        self.__lifetime: float = float(lifetime)
        self.__uptime: float = time()
        self.__data: CADT = data

    def __str__(self) -> str:
        return f"cache object at {id(self)}"

    @property
    def up(self) -> float:
        '''uptime'''
        return self.__uptime

    @property
    def age(self) -> float:
        '''runtime'''
        return time() - self.up

    @property
    def life(self) -> float:
        '''lifetime'''
        return self.__lifetime

    @property
    def down(self) -> float:
        '''downtime'''
        return self.life - self.age if self.life > 0.0 else 0.0

    @property
    def expired(self) -> bool:
        return self.life > 0.0 and self.age > self.life

    def renew(self, lifetime: Optional[CacheTimeUnit] = None) -> None:
        '''renew uptime and lifetime(optional)'''
        if lifetime is not None:
            self.__lifetime = float(lifetime)
        self.__uptime = time()

    def update(self, data: CADT) -> None:
        '''update cache data'''
        self.__data = data
        self.renew()

    @property
    def data(self) -> CADT:
        return self.__data

    @data.setter
    def data(self, data: CADT) -> None:
        self.update(data)


CDT = TypeVar("CDT")


class CacheData(CacheAtom[CDT]):
    '''Data cache with enforces expiration check'''

    @property
    def data(self) -> CDT:
        if self.expired:
            raise CacheExpired()
        return super().data

    @data.setter
    def data(self, data: CDT) -> None:
        super().update(data)


NCNT = TypeVar("NCNT")
NCDT = TypeVar("NCDT")


class NamedCache(CacheAtom[NCDT], Generic[NCNT, NCDT]):
    '''Named data cache'''

    def __init__(self, name: NCNT, data: NCDT, lifetime: CacheTimeUnit = 0):
        super().__init__(data, lifetime)
        self.__name: NCNT = name

    def __str__(self) -> str:
        return f"cache object at {id(self)} name={self.name}"

    @property
    def name(self) -> NCNT:
        return self.__name


CINT = TypeVar("CINT")
CIDT = TypeVar("CIDT")


class CacheItem(NamedCache[CINT, CIDT]):
    '''Named data cache with enforces expiration check'''

    def __init__(self, name: CINT, data: CIDT, lifetime: CacheTimeUnit = 0):
        super().__init__(name, data, lifetime)

    @property
    def data(self) -> CIDT:
        if self.expired:
            raise CacheExpired(self.name)
        return super().data

    @data.setter
    def data(self, data: CIDT) -> None:
        super().update(data)


IPKT = TypeVar("IPKT")
IPVT = TypeVar("IPVT")


class ItemPool(Generic[IPKT, IPVT]):
    '''Cache item pool'''

    def __init__(self, lifetime: CacheTimeUnit = 0):
        self.__pool: Dict[IPKT, CacheItem[IPKT, IPVT]] = {}
        self.__lifetime: float = float(lifetime)
        self.__intlock: Lock = Lock()  # internal lock

    def __str__(self) -> str:
        return f"cache item pool at {id(self)}"

    def __len__(self) -> int:
        with self.__intlock:
            return len(self.__pool)

    def __iter__(self) -> Iterator[IPKT]:
        with self.__intlock:
            return iter(self.__pool.keys())

    def __contains__(self, index: IPKT) -> bool:
        with self.__intlock:
            return index in self.__pool

    def __setitem__(self, index: IPKT, value: IPVT) -> None:
        return self.put(index, value)

    def __getitem__(self, index: IPKT) -> CacheItem:
        return self.get(index)

    def __delitem__(self, index: IPKT) -> None:
        return self.delete(index)

    @property
    def lifetime(self) -> float:
        return self.__lifetime

    @lifetime.setter
    def lifetime(self, lifetime: CacheTimeUnit) -> None:
        self.__lifetime = float(lifetime)

    def put(self, index: IPKT, value: IPVT, lifetime: Optional[CacheTimeUnit] = None) -> None:  # noqa:E501
        life = lifetime if lifetime is not None else self.lifetime
        item = CacheItem(index, value, life)
        with self.__intlock:
            self.__pool[index] = item

    def get(self, index: IPKT) -> CacheItem[IPKT, IPVT]:
        with self.__intlock:
            try:
                return self.__pool[index]
            except KeyError as exc:
                raise CacheMiss(index) from exc

    def delete(self, index: IPKT) -> None:
        with self.__intlock:
            if index in self.__pool:
                del self.__pool[index]


CPIT = TypeVar("CPIT")
CPVT = TypeVar("CPVT")


class CachePool(ItemPool[CPIT, CPVT]):
    '''Named data cache pool'''

    def __init__(self, lifetime: CacheTimeUnit = 0):
        super().__init__(lifetime=lifetime)

    def __str__(self) -> str:
        return f"cache pool at {id(self)}"

    def __getitem__(self, index: CPIT) -> CPVT:
        return self.get(index)

    def get(self, index: CPIT) -> CPVT:
        try:
            return super().get(index).data
        except CacheExpired as exc:
            super().delete(index)
            assert index not in self
            raise CacheMiss(index) from exc



# coding:utf-8

from enum import Enum
import logging
import os
import sys
from time import time
from typing import Iterable
from typing import List
from typing import Optional
from typing import Set
from typing import TextIO

from colorlog import ColoredFormatter
from colorlog.formatter import LogColors

from xkits.attribute import __prog_name__ as prog


class once_filter(logging.Filter):
    def __init__(self, name: str = "logging_once_filter",
                 max_interval_seconds: int = 60):
        self.__max_interval: int = max(3, max_interval_seconds)
        self.__timestamp: float = time()
        self.__message = None
        super().__init__(name)

    def filter(self, record: logging.LogRecord) -> bool:
        current_message = (record.msg, record.args)
        if current_message != self.__message or self.timeout:
            self.__message = current_message
            self.__timestamp = time()
            return True
        return False

    @property
    def max_interval_seconds(self) -> int:
        return self.__max_interval

    @property
    def interval_seconds(self) -> float:
        return time() - self.__timestamp

    @property
    def timeout(self) -> bool:
        return self.interval_seconds > self.max_interval_seconds


class log:
    class LOG_LEVELS(Enum):
        CRITICAL = "fatal"
        ERROR = "error"
        WARNING = "warn"
        INFO = "info"
        DEBUG = "debug"
    ALLOWED_LOG_LEVELS: List[str] = [
        LOG_LEVELS.CRITICAL.value,
        LOG_LEVELS.ERROR.value,
        LOG_LEVELS.WARNING.value,
        LOG_LEVELS.INFO.value,
        LOG_LEVELS.DEBUG.value,
    ]
    DEFAULT_LOG_FORMAT: str = "%(log_color)s%(message)s"
    DEFAULT_LOG_COLORS: LogColors = {
        LOG_LEVELS.CRITICAL.name: "light_red",
        LOG_LEVELS.ERROR.name: "red",
        LOG_LEVELS.WARNING.name: "yellow",
        LOG_LEVELS.INFO.name: "white",
        LOG_LEVELS.DEBUG.name: "black",
    }

    def __init__(self):
        self.__initiated_names: Set[str] = set()

    def logger_initiated(self, name: str) -> bool:
        return name in self.__initiated_names

    def get_logger(self, name: Optional[str] = None) -> logging.Logger:
        return logging.getLogger(name if isinstance(name, str) else prog)

    def initiate_logger(self, logger: logging.Logger,
                        level: Optional[str] = None,
                        handlers: Optional[Iterable[logging.Handler]] = None,
                        filters: Optional[Iterable[logging.Filter]] = None):
        if self.logger_initiated(logger.name):
            logger.warning(F"logger {logger.name} is already initiated")

        if isinstance(level, str):
            name = level.upper()
            logger.setLevel(
                logging._nameToLevel[name]  # pylint: disable=protected-access
            )

        if filters is None:
            filters = [once_filter()]

        for _filter in filters:
            logger.addFilter(_filter)

        if handlers is None:
            handlers = [self.new_stream_handler(stream=sys.stdout)]

        for handler in handlers:
            logger.addHandler(handler)

        self.__initiated_names.add(logger.name)

    @classmethod
    def new_stream_handler(cls, stream: Optional[TextIO] = None,
                           fmt: Optional[str] = None,
                           log_colors: Optional[LogColors] = None
                           ) -> logging.StreamHandler:
        if fmt is None:
            fmt = cls.DEFAULT_LOG_FORMAT
        if log_colors is None:
            log_colors = cls.DEFAULT_LOG_COLORS
        formatter: logging.Formatter = ColoredFormatter(
            fmt=fmt, datefmt=None, log_colors=log_colors)
        handler: logging.StreamHandler = logging.StreamHandler(stream=stream)
        handler.setFormatter(formatter)
        return handler

    @classmethod
    def new_file_handler(cls, filename: str,
                         fmt: Optional[str] = None) -> logging.FileHandler:
        assert isinstance(filename, str), f"Unexpected type: {type(filename)}"
        dirname: str = os.path.dirname(filename)
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        assert os.path.isdir(dirname), f"{dirname} not an existing directory"
        formatter: logging.Formatter = ColoredFormatter(
            fmt=fmt, datefmt=None, no_color=True)
        handler: logging.FileHandler = logging.FileHandler(filename=filename)
        handler.setFormatter(formatter)
        return handler



# coding:utf-8

from base64 import b64encode
from datetime import datetime
from enum import Enum
import os
from typing import Dict
from typing import Optional
from typing import Union
from urllib.parse import ParseResult
from urllib.parse import urljoin
from urllib.parse import urlparse
from urllib.parse import urlunparse

from bs4 import BeautifulSoup
from requests import Response
from requests import Session

from xkits.cache import CacheMiss
from xkits.cache import CachePool
from xkits.cache import CacheTimeUnit

SessionTimeUnit = Union[float, int]
SessionTimeout = Optional[SessionTimeUnit]


class ProxyProtocol(Enum):
    HTTP = "http"
    HTTPS = "https"
    SOCKS4 = "socks4"
    SOCKS5 = "socks5"


class ProxySession(Session):

    def __init__(self, protocol: ProxyProtocol, host: str, port: int):
        if not isinstance(protocol, ProxyProtocol):
            raise ValueError(f"Invalid proxy protocol: {protocol}")
        proxies = {
            "http": f"{protocol.value}://{host}:{port}",
            "https": f"{protocol.value}://{host}:{port}",
        }
        super(ProxySession, self).__init__()  # pylint: disable=R1725
        self.proxies.update(proxies)

    @classmethod
    def http_proxy(cls, host: str, port: int = 80) -> "ProxySession":
        return cls(ProxyProtocol.HTTP, host, port)

    @classmethod
    def https_proxy(cls, host: str, port: int = 443) -> "ProxySession":
        return cls(ProxyProtocol.HTTPS, host, port)

    @classmethod
    def socks4_proxy(cls, host: str, port: int = 1080) -> "ProxySession":
        return cls(ProxyProtocol.SOCKS4, host, port)

    @classmethod
    def socks5_proxy(cls, host: str, port: int = 1080) -> "ProxySession":
        return cls(ProxyProtocol.SOCKS5, host, port)


class PageConnect(object):  # pylint: disable=useless-object-inheritance
    '''Website page without cached response'''

    def __init__(self, url: str, session: Session, timeout: SessionTimeout = None):  # noqa:E501
        self.__timeout: SessionTimeout = timeout
        self.__session: Session = session
        self.__url: str = url

    def __str__(self) -> str:
        return f"page object at {id(self)} url={self.url}"

    @property
    def url(self) -> str:
        return self.__url

    @property
    def session(self) -> Session:
        return self.__session

    @property
    def timeout(self) -> SessionTimeout:
        return self.__timeout

    def get(self, timeout: SessionTimeout = None) -> Response:
        response = self.session.get(self.url, timeout=timeout or self.timeout)
        response.raise_for_status()
        return response


class Page(PageConnect):
    '''Website page with cached response'''

    def __init__(self, url: str, session: Optional[Session] = None, timeout: SessionTimeout = None):  # noqa:E501
        super().__init__(url=url, session=session or Session(), timeout=timeout)  # noqa:E501
        self.__response: Optional[Response] = None

    @property
    def label(self) -> str:
        encode: bytes = self.url.encode(encoding="utf-8")
        decode: str = b64encode(encode).decode(encoding="utf-8").rstrip("=")
        return f"{datetime.now().strftime('%Y%m%d%H%M%S')}-{decode}"

    @property
    def response(self) -> Response:
        if self.__response is None:
            self.__response = self.get()
        return self.__response

    @property
    def soup(self) -> BeautifulSoup:
        return BeautifulSoup(self.response.content, "html.parser")

    def save(self, path: Optional[str] = None) -> str:
        file: str = self.label if path is None else os.path.join(path, self.label) if os.path.isdir(path) else path  # noqa:E501
        with open(file=file, mode="wb") as hdl:
            hdl.write(self.response.content)
        return file


class PageCache(CachePool[str, Page]):
    '''Website pages cache pool'''

    def __init__(self, session: Optional[Session] = None,
                 lifetime: CacheTimeUnit = 0):
        self.__session: Optional[Session] = session
        super().__init__(lifetime=lifetime)

    def __str__(self) -> str:
        return f"website pages cache pool at {id(self)}"  # pragma: no cover

    def __getitem__(self, url: str) -> Page:
        return self.fetch(url=url)

    @property
    def session(self) -> Session:
        return self.__session or Session()

    def fetch(self, url: str, session: Optional[Session] = None,
              timeout: SessionTimeout = None) -> Page:
        while True:
            try:
                return super().get(url)
            except CacheMiss:
                page = Page(url=url, session=session or self.session,
                            timeout=timeout)
                super().put(url, page)


class Site(PageCache):
    '''Website with pages cache pool'''

    def __init__(self, base: str, session: Optional[Session] = None,
                 lifetime: CacheTimeUnit = 0):
        super().__init__(session=session, lifetime=lifetime)
        components: ParseResult = urlparse(url=base)
        self.__baseurl: str = urlunparse(components)
        self.__components: ParseResult = components

    def __str__(self) -> str:
        return f"website {self.base} with pages cache pool"

    @property
    def scheme(self) -> str:
        return self.__components.scheme

    @property
    def netloc(self) -> str:
        return self.__components.netloc

    @property
    def main(self) -> str:
        '''main page url'''
        return urlunparse((self.scheme, self.netloc, '', '', '', ''))

    @property
    def base(self) -> str:
        '''base url'''
        return self.__baseurl

    def parse(self, *path: str) -> str:
        '''parse page url'''
        site: str = "/".join([self.__components.path] + list(path))
        return urljoin(base=self.main, url=site)

    def login(self, url: str, data: Dict[str, str]) -> Response:
        response = self.session.post(url=url, data=data)
        return response

    def page(self, *path: str, timeout: SessionTimeout = None) -> Page:
        return self.fetch(url=self.parse(*path), session=self.session, timeout=timeout)  # noqa:E501
